import unittest

from dsplib.models.connection import Connection
from dsplib.models.helpers import BaseError, Actions, LastModificationDate
from dsplib.models.ontology import Ontology


class TestOntology(unittest.TestCase):
    project = "http://rdfh.ch/projects/0001"
    label = 'test_ontology'
    last_modification_date: LastModificationDate

    def setUp(self) -> None:
        self.con = Connection('http://0.0.0.0:3333')
        self.con.login('root@example.com', 'test')

    def test_ontology(self):
        onto = Ontology(
            con=self.con,
            project=self.project,
            name="test_ontology_0",
            label="test ontology 0",
            lastModificationDate="2017-12-19T15:23:42.166Z"
        )
        self.assertEqual(onto.project, self.project)
        self.assertEqual(onto.label, "test ontology 0")
        self.assertEqual(onto.lastModificationDate, LastModificationDate("2017-12-19T15:23:42.166Z"))

    def test_read(self):
        onto = Ontology(
            con=self.con,
            id='http://0.0.0.0:3333/ontology/0001/anything/v2'
        ).read()
        self.assertEqual(onto.id, "http://0.0.0.0:3333/ontology/0001/anything/v2")
        self.assertEqual(onto.project, "http://rdfh.ch/projects/0001")
        self.assertEqual(onto.label, "The anything ontology")
        self.assertIsNotNone(onto.lastModificationDate)

    def test_create(self):
        onto = Ontology(
            con=self.con,
            project=self.project,
            name="testonto_1",
            label="test ontology 1",
        ).create()
        self.assertIsNotNone(onto.id)
        self.assertEqual(onto.id, 'http://0.0.0.0:3333/ontology/0001/' + "testonto_1" + '/v2')
        self.assertEqual(onto.label, "test ontology 1")
        self.assertEqual(onto.project, self.project)

    def test_update(self):
        onto = Ontology(
            con=self.con,
            project=self.project,
            name="testonto_2",
            label="test ontology 2",
        ).create()
        self.assertIsNotNone(onto.id)
        onto.label = 'This is a modified label!'
        onto = onto.update()
        self.assertEqual(onto.label, 'This is a modified label!')

    def test_delete(self):
        onto = Ontology(
            con=self.con,
            project=self.project,
            name="testonto_3",
            label="test ontology 3",
        ).create()
        self.assertIsNotNone(onto.id)
        res = onto.delete()
        self.assertIsNotNone(res)

    def test_get_ontologies_of_project(self):
        ontolist = Ontology.getProjectOntologies(self.con, self.project)
        ontolist_ids = [x.id for x in ontolist]
        self.assertIn('http://0.0.0.0:3333/ontology/0001/anything/v2', ontolist_ids)
        self.assertIn('http://0.0.0.0:3333/ontology/0001/minimal/v2', ontolist_ids)
        self.assertIn('http://0.0.0.0:3333/ontology/0001/something/v2', ontolist_ids)


if __name__ == '__main__':
    unittest.main()



"""
This module handles the import of XML data into the DSP platform.
"""
import os
from typing import Dict, List, Optional, Union

from lxml import etree

from dsplib.models.connection import Connection
from dsplib.models.group import Group
from dsplib.models.permission import Permissions
from dsplib.models.project import Project
from dsplib.models.resource import ResourceInstanceFactory
from dsplib.models.sipi import Sipi
from dsplib.models.value import KnoraStandoffXml


class XmlError(BaseException):
    """Represents an error raised in the context of the XML import"""
    _message: str

    def __init__(self, msg: str):
        self._message = msg

    def __str__(self):
        return 'XML-ERROR: ' + self._message


class ProjectContext:
    """Represents the project context"""

    _projects: list[Project]
    _project_map: Dict[str, str]  # dictionary of (project name:project IRI) pairs
    _inv_project_map: Dict[str, str]  # dictionary of (project IRI:project name) pairs
    _groups: list[Group]
    _group_map: Dict[str, str]
    _shortcode: Optional[str]
    _project_name: Optional[str]

    def __init__(self, con: Connection, shortcode: Optional[str] = None):
        self._shortcode = shortcode
        self._projects = Project.getAllProjects(con=con)
        self._project_map: Dict[str, str] = {x.shortname: x.id for x in self._projects}
        self._inv_project_map: Dict[str, str] = {x.id: x.shortname for x in self._projects}
        self._groups = Group.getAllGroups(con=con)
        self._group_map: Dict[str, str] = {self._inv_project_map[x.project] + ':' + x.name: x.id for x in self._groups}
        self._project_name = None
        # get the project name from the shortcode
        if self._shortcode:
            for p in self._projects:
                if p.shortcode == self._shortcode:
                    self._project_name = p.shortname
                    break

    @property
    def group_map(self) -> Dict[str, str]:
        """Dictionary of (project:group name) and (group id) pairs of all groups in project"""
        return self._group_map

    @property
    def project_name(self) -> Optional[str]:
        """Name of the project"""
        return self._project_name


class KnoraValue:
    """Represents a value of a resource in the Knora ontology"""

    _value: Union[str, KnoraStandoffXml]
    _resrefs: Optional[List[str]]
    _comment: str
    _permissions: str
    _is_richtext: bool

    def __init__(self, node: etree.Element, val_type: str, listname: Optional[str] = None) -> None:

        self._resrefs = None
        self._comment = node.get('comment')
        self._permissions = node.get('permissions')
        if node.get('encoding') == 'xml':
            node.attrib.clear()
            xmlstr = etree.tostring(node, encoding="unicode", method="xml")
            xmlstr = xmlstr.replace('<text xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">', '')
            xmlstr = xmlstr.replace('</text>', '')
            self._value = KnoraStandoffXml(xmlstr)
            tmp_id_list = self._value.findall()
            if tmp_id_list:
                refs = set()
                for tmp_id in tmp_id_list:
                    refs.add(tmp_id.split(':')[1])
                self._resrefs = list(refs)
        else:
            if val_type == 'list':
                self._value = listname + ':' + "".join(node.itertext())
            else:
                self._value = "".join(node.itertext())

    @property
    def value(self) -> Union[str, KnoraStandoffXml]:
        """The actual value of the value instance"""
        return self._value

    @property
    def resrefs(self) -> Optional[List[str]]:
        """List of resource references"""
        return self._resrefs

    @property
    def comment(self) -> str:
        """Comment about the value"""
        return self._comment

    @property
    def permissions(self) -> str:
        """Reference to the set of permissions for the value"""
        return self._permissions

    @property
    def is_richtext(self) -> bool:
        """true if text value is of type richtext, false otherwise"""
        return self._is_richtext

    def print(self) -> None:
        """Prints the value and its attributes."""
        print('   Value: ' + str(self._value))
        if self._comment:
            print('   Comment:' + self._comment)
        if self._resrefs is not None:
            for i in self._resrefs:
                print('   res_ref: ' + i)


class KnoraProperty:
    """Represents a property of a resource in the XML"""

    _name: str
    _valtype: str
    _values: List[KnoraValue]

    def __init__(self, node: etree.Element, valtype: str, default_ontology: Optional[str] = None):
        """
        The constructor for the knora property

        Args:
            node: the property node, p.ex. <decimal-prop></decimal-prop>
            valtype: the type of value given by the name of the property node, p.ex. decimal in <decimal-prop>
            default_ontology: the name of the ontology
        """
        # get the property name which is in format namespace:propertyname, p.ex. rosetta:hasName
        tmp_prop_name = node.attrib['name'].split(':')
        if len(tmp_prop_name) > 1:
            if tmp_prop_name[0]:
                self._name = node.attrib['name']
            else:
                # replace an empty namespace with the default ontology name
                self._name = default_ontology + ':' + tmp_prop_name[1]
        else:
            self._name = 'knora-admin:' + tmp_prop_name[0]
        listname = node.attrib.get('list')  # safe the list name if given (only for lists)
        self._valtype = valtype
        self._values = []

        # parse the subnodes of the property nodes which contain the actual values of the property
        for subnode in node:
            if subnode.tag == valtype:  # the subnode must correspond to the expected value type
                self._values.append(KnoraValue(subnode, valtype, listname))
            else:
                raise XmlError('Unexpected tag: "{}". Property may contain only value tags!'.format(subnode.tag))

    @property
    def name(self) -> str:
        """The name of the property"""
        return self._name

    @property
    def valtype(self) -> str:
        """The value type of the property"""
        return self._valtype

    @property
    def values(self) -> List[KnoraValue]:
        """List of values of this property"""
        return self._values

    def print(self) -> None:
        """Prints the property."""
        print('  Property: {} Type: {}'.format(self._name, self._valtype))
        for value in self._values:
            value.print()


class KnoraResource:
    """Represents a resource in the Knora ontology"""

    _id: str
    _label: str
    _restype: str
    _permissions: str
    _bitstream: Optional[str]
    _properties: List[KnoraProperty]

    def __init__(self, node: etree.Element, default_ontology: Optional[str] = None) -> None:
        """
        Constructor that parses a resource node from the XML DOM

        Args:
            node: The DOM node to be processed representing a resource (which is a child of the knora element)
            default_ontology: The default ontology (given in the attribute default-ontology of the knora element)
        """
        self._id = node.attrib['id']
        self._label = node.attrib['label']
        # get the resource type which is in format namespace:resourcetype, p.ex. rosetta:Image
        tmp_res_type = node.attrib['restype'].split(':')
        if len(tmp_res_type) > 1:
            if tmp_res_type[0]:
                self._restype = node.attrib['restype']
            else:
                # replace an empty namespace with the default ontology name
                self._restype = default_ontology + ':' + tmp_res_type[1]
        else:
            self._restype = 'knora-admin:' + tmp_res_type[0]
        self._permissions = node.attrib['permissions']
        self._bitstream = None
        self._properties = []
        for subnode in node:
            if subnode.tag == 'bitstream':
                self._bitstream = subnode.text  # path to the file
            elif subnode.tag is etree.Comment:
                continue
            else:
                # get the property type which is in format type-prop, p.ex. <decimal-prop>
                prop_type, _ = subnode.tag.split('-')
                self._properties.append(KnoraProperty(subnode, prop_type, default_ontology))

    @property
    def id(self) -> str:
        """The unique id of the resource"""
        return self._id

    @property
    def label(self) -> str:
        """The label of the resource"""
        return self._label

    @property
    def restype(self) -> str:
        """The type of the resource"""
        return self._restype

    @property
    def permissions(self) -> str:
        """The reference to the permissions set for this resource"""
        return self._permissions

    @property
    def bitstream(self) -> Optional[str]:
        """The path to the bitstream object (file) belonging to the resource"""
        return self._bitstream

    def print(self) -> None:
        """Prints the resource and its attributes."""
        print(f'Resource: id={self._id}, restype: {self._restype}, label: {self._label}')
        if self._bitstream is not None:
            print('  Bitstream: ' + self._bitstream)
        for prop in self._properties:
            prop.print()

    def get_resptrs(self) -> List[str]:
        """
        Get a list of all resource id's that are referenced by this resource

        Returns:
            List of resources identified by their unique id's (as given in the XML)
        """
        resptrs: List[str] = []
        for prop in self._properties:
            if prop.valtype == 'resptr':
                for value in prop.values:
                    resptrs.append(value.value)
            elif prop.valtype == 'text':
                for value in prop.values:
                    if value.resrefs is not None:
                        resptrs.extend(value.resrefs)
        return resptrs

    def get_propvals(self, resiri_lookup: Dict[str, str], permissions_lookup: Dict[str, Permissions]) -> Dict[str, Permissions]:
        """
        Get a dictionary of the property names and their values belonging to a resource

        Args:
            resiri_lookup: Is used to solve internal unique id's of resources to real IRI's
            permissions_lookup: Is used to resolve the permission id's to permission sets

        Returns:
            A dict of values with the property name as key and a single value. This dict represents the JSON structure
            that Knora.create_resource() expects.
        """
        prop_data = {}
        for prop in self._properties:
            vals: List[Union[str, Dict[str, str]]] = []
            for value in prop.values:
                v: str
                if prop.valtype == 'resptr':  # we have a resptr, therefore simple lookup or IRI
                    iri = resiri_lookup.get(value.value)
                    if iri is not None:
                        v = iri
                    else:
                        v = value.value  # if we do not find the unique_id, we assume it's a valid knora IRI
                elif prop.valtype == 'text':
                    if isinstance(value.value, KnoraStandoffXml):
                        iri_refs = value.value.findall()  # The IRI's must be embedded  as "...IRI:unique_id:IRI..."
                        for iri_ref in iri_refs:
                            res_id = iri_ref.split(':')[1]
                            iri = resiri_lookup.get(res_id)
                            value.value.replace(iri_ref, iri)
                    v = value.value
                else:
                    v = value.value

                if value.comment is None and value.permissions is None:
                    # no comment or permissions
                    vals.append(v)
                else:
                    # we have comment or permissions
                    tmp = {'value': v}
                    if value.comment is not None:
                        tmp['comment'] = value.comment
                    if value.permissions is not None:
                        tmp['permissions'] = permissions_lookup.get(value.permissions)
                    vals.append(tmp)
            prop_data[prop.name] = vals if len(vals) > 1 else vals[0]
        return prop_data


class XmlAllow:
    """Represents the allow element of the XML"""

    _group: str
    _permission: str

    def __init__(self, node: etree.Element, project_context: ProjectContext) -> None:
        """
        Constructor which parses the XML DOM allow element

        Args:
            node: The DOM node to be processed (represents a single right in a permission set)
            project_context: Context for DOM node traversal

        Returns:
            None
        """
        tmp = node.attrib['group'].split(':')
        sysgroups = ['UnknownUser', 'KnownUser', 'ProjectMember', 'Creator', 'ProjectAdmin', 'SystemAdmin']
        if len(tmp) > 1:
            if tmp[0]:
                if tmp[0] == 'knora-admin' and tmp[1] in sysgroups:
                    self._group = node.attrib['group']
                else:
                    self._group = project_context.group_map.get(node.attrib['group'])
                    if self._group is None:
                        raise XmlError("Group \"{}\" is not known: Cannot find project!".format(node.attrib['group']))
            else:
                if project_context.project_name is None:
                    raise XmlError("Project shortcode has not been set in ProjectContext")
                self._group = project_context.project_name + ':' + tmp[1]
        else:
            if tmp[0] in sysgroups:
                self._group = 'knora-admin:' + node.attrib['group']
            else:
                raise XmlError("Group \"{}\" is not known: ".format(node.attrib['group']))
        self._permission = node.text

    @property
    def group(self) -> str:
        """The group specified in the allow element"""
        return self._group

    @property
    def permission(self) -> str:
        """The reference to a set of permissions"""
        return self._permission

    def print(self) -> None:
        """Prints the attributes of the XmlAllow instance"""
        print("  group=", self._group, " permission=", self._permission)


class XmlPermission:
    """Represents the permission set containing several XmlAllow elements"""

    _id: str
    _allows: List[XmlAllow]

    def __init__(self, node: etree.Element, project_context: ProjectContext) -> None:
        """
        Constructor which parses a XML DOM permissions element representing an named permission set

        Args:
            node: The DOM node to be processed (representing an a permission set)
            project_context: Context for DOM node traversal
        """
        self._allows = []
        self._id = node.attrib['id']
        for allow_node in node:
            self._allows.append(XmlAllow(allow_node, project_context))

    @property
    def id(self) -> str:
        """The id of the permission set, p.ex. res-default"""
        return self._id

    @property
    def allows(self) -> List[XmlAllow]:
        """List of XmlAllow elements defining permissions for specific groups"""
        return self._allows

    def get_permission_instance(self) -> Permissions:
        """Returns a list of allow elements of this permission instance"""
        permissions = Permissions()
        for allow in self._allows:
            permissions.add(allow.permission, allow.group)
        return permissions

    def __str__(self):
        allow_str: List[str] = []
        for allow in self._allows:
            allow_str.append("{} {}".format(allow.permission, allow.group))
        return '|'.join(allow_str)

    def print(self):
        """Prints the permission set"""
        print('Permission: ', self._id)
        for a in self._allows:
            a.print()


def do_sort_order(resources: List[KnoraResource]) -> List[KnoraResource]:
    """
    Sorts a list of resources.

    The sorting is such that resources that reference other resources are added after the referenced resources. It
    will fail with an error if there are circular references.

    Args:
        resources: List of resources before sorting

    Returns:
        sorted list of resources
    """

    # sort the resources according to outgoing resptrs
    ok_resources: [KnoraResource] = []
    notok_resources: [KnoraResource] = []
    ok_res_ids: [str] = []
    cnt = 0
    notok_len = 9999999
    while len(resources) > 0 and cnt < 10000:
        for resource in resources:
            resptrs = resource.get_resptrs()
            if len(resptrs) == 0:
                ok_resources.append(resource)
                ok_res_ids.append(resource.id)
            else:
                ok = True
                for resptr in resptrs:
                    if resptr in ok_res_ids:
                        pass
                    else:
                        ok = False
                if ok:
                    ok_resources.append(resource)
                    ok_res_ids.append(resource.id)
                else:
                    notok_resources.append(resource)
        resources = notok_resources
        if not len(notok_resources) < notok_len:
            print('Cannot resolve resptr dependencies. Giving up....')
            print(len(notok_resources))
            for r in notok_resources:
                print('Resource {} has unresolvable resptrs to: '.format(r.id), end=' ')
                for x in r.get_resptrs():
                    print(x, end=' ')
                print('')
                print('=============')
            exit(5)
        notok_len = len(notok_resources)
        notok_resources = []
        cnt += 1
        print('{}. Ordering pass Finished!'.format(cnt))
    # print('Remaining: {}'.format(len(resources)))
    return ok_resources


def validate_xml_against_schema(input_file: str, schema_file: str) -> bool:
    """
    Validates an XML file against an XSD schema

    Args:
        input_file: the XML file to be validated
        schema_file: the schema against which the XML file should be validated

    Returns:
        True if the XML file is valid, False otherwise
    """
    xmlschema = etree.XMLSchema(schema_file)
    doc = etree.parse(input_file)

    is_valid = False

    if xmlschema.assertValid(doc):
        is_valid = True

    return is_valid


def xml_upload(input_file: str, server: str, user: str, password: str, imgdir: str, sipi: str, verbose: bool,
               validate_only: bool) -> bool:
    """
    This function reads an XML file and imports the data described in it onto the DSP server.

    Args:
        input_file : the XML with the data to be imported onto the DSP server
        server : the DSP server where the data should be imported
        user : the user (e-mail) with which the data should be imported
        password : the password of the user with which the data should be imported
        imgdir : the image directory
        sipi : the sipi instance to be used
        verbose : verbose option for the command, if used more output is given to the user
        validate_only : validation option to validate the XML data without the actual import of the data

    Returns:
        None
    """

    # Validate the input XML file
    current_dir = os.path.dirname(os.path.realpath(__file__))
    schema_file = etree.parse(os.path.join(current_dir, '../schemas/data.xsd'))

    if validate_xml_against_schema(input_file, schema_file):
        print("The input data file is syntactically correct and passed validation!")
        if validate_only:
            return True

    # Connect to the DaSCH Service Platform API and get the project context
    con = Connection(server)
    con.login(user, password)
    proj_context = ProjectContext(con=con)

    resources: List[KnoraResource] = []
    permissions: Dict[str, XmlPermission] = {}

    # parse the XML file containing the data
    tree = etree.parse(input_file)
    knora = tree.getroot()
    default_ontology = knora.attrib['default-ontology']
    shortcode = knora.attrib['shortcode']

    for child in knora:
        # get all permissions
        if child.tag == "permissions":
            permission = XmlPermission(child, proj_context)
            permissions[permission.id] = permission
        # get all resources
        elif child.tag == "resource":
            resources.append(KnoraResource(child, default_ontology))

    # sort the resources (resources which do not link to others come first)
    resources = do_sort_order(resources)

    sipi = Sipi(sipi, con.get_token())

    # get the project information and project ontology from the server
    project = ResourceInstanceFactory(con, shortcode)

    # create a dictionary to look up permissions
    permissions_lookup: Dict[str, Permissions] = {}
    for key, perm in permissions.items():
        permissions_lookup[key] = perm.get_permission_instance()

    # create a dictionary to look up resource classes
    res_classes: Dict[str, type] = {}
    for res_class_name in project.get_resclass_names():
        res_classes[res_class_name] = project.get_resclass(res_class_name)

    res_iri_lookup: Dict[str, str] = {}

    for resource in resources:
        if verbose:
            resource.print()
        if resource.bitstream:
            img = sipi.upload_bitstream(os.path.join(imgdir, resource.bitstream))
            bitstream = img['uploadedFiles'][0]['internalFilename']
        else:
            bitstream = None

        # create the resource on the server
        instance = res_classes[resource.restype](con=con, label=resource.label,
                                                 permissions=permissions_lookup.get(resource.permissions), bitstream=bitstream,
                                                 values=resource.get_propvals(res_iri_lookup, permissions_lookup)).create()
        res_iri_lookup[resource.id] = instance.iri
        print("Created resource: ", instance.label, " (", resource.id, ") with IRI ", instance.iri)



from typing import Dict, List

from dsplib.utils.excel_to_json_lists import make_json_list_from_excel, prepare_list_creation


def expand_lists_from_excel(data_model: Dict) -> List[str]:
    """
    Gets all lists from an ontology and expands them to json if they are only referenced via an Excel file

    Args:
        data_model: The data model (json) the lists are read from

    Returns:
        A list of all expanded lists. It can be added to the root node of an ontology as list section.
    """
    # create and add lists from Excel references to the ontology
    lists = data_model['project'].get('lists')
    new_lists = []
    if lists is not None:
        for rootnode in lists:
            # check if the folder parameter is used
            if rootnode.get('nodes') is not None and isinstance(rootnode['nodes'], dict) and rootnode['nodes'].get(
                    'folder') is not None:
                # get the Excel files from the folder and crate the rootnode of the list
                excel_folder = rootnode['nodes']['folder']
                rootnode, excel_files = prepare_list_creation(excel_folder, rootnode.get('name'))

                # create the list from the Excel files
                make_json_list_from_excel(rootnode, excel_files)

                new_lists.append(rootnode)
            else:
                new_lists.append(rootnode)

    return new_lists



import unittest
from pprint import pprint
from knora.dsplib.models.value import DateValue
from knora.dsplib.models.helpers import Actions

class TestValue(unittest.TestCase):

    def test_date_value(self):
        date: DateValue = DateValue("JULIAN:BCE:0700:BCE:0600")
        self.assertEqual(date._calendar, 'JULIAN')

        self.assertEqual(date._e1, 'BCE')
        self.assertEqual(date._y1, 700)
        self.assertEqual(date._m1, None)
        self.assertEqual(date._d1, None)

        self.assertEqual(date._e2, 'BCE')
        self.assertEqual(date._y2, 600)
        self.assertEqual(date._m2, None)
        self.assertEqual(date._d2, None)

        jsonLD = date.toJsonLdObj(Actions.Create)

        self.assertEqual(jsonLD['knora-api:dateValueHasCalendar'], 'JULIAN')
        self.assertEqual(jsonLD['knora-api:dateValueHasStartEra'], 'BCE')
        self.assertEqual(jsonLD['knora-api:dateValueHasStartYear'], 700)
        self.assertEqual(jsonLD['knora-api:dateValueHasEndYear'], 600)
        self.assertEqual(jsonLD['knora-api:dateValueHasEndEra'], 'BCE')


if __name__ == '__main__':
    unittest.main()



import json
import os

import jsonschema
from openpyxl import load_workbook


def validate_resources_with_schema(json_file: str) -> bool:
    """
        This function checks if the json resources are valid according to the schema.

        Args:
            json_file: the json with the resources to be validated

        Returns:
            True if the data passed validation, False otherwise

        """
    current_dir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(current_dir, '../schemas/resources-only.json')) as schema:
        resources_schema = json.load(schema)

    try:
        jsonschema.validate(instance=json_file, schema=resources_schema)
    except jsonschema.exceptions.ValidationError as err:
        print(err)
        return False
    print('Resource data passed schema validation.')
    return True


def resources_excel2json(excelfile: str, outfile: str):
    """
        Converts properties described in an Excel file into a properties section which can be integrated into a DSP ontology

        Args:
            excelfile: path to the Excel file containing the properties
            outfile: path to the output JSON file containing the properties section for the ontology

        Returns:
            None
    """
    # load file
    wb = load_workbook(excelfile, read_only=True)

    # get overview
    sheet = wb['classes']
    resource_list = [c for c in sheet.iter_rows(min_row=2, values_only=True)]

    prefix = '"resources":'
    resources = []
    # for each resource in resources overview
    for res in resource_list:
        # get name
        name = res[0]
        # get labels
        labels = {}
        if res[1]:
            labels['en'] = res[1]
        if res[2]:
            labels['de'] = res[2]
        if res[3]:
            labels['fr'] = res[3]
        if res[4]:
            labels['it'] = res[4]
        # get super
        sup = res[5]

        # load details for this resource
        sh = wb[name]
        property_list = [c for c in sh.iter_rows(min_row=2, values_only=True)]

        cards = []
        # for each of the detail sheets
        for i, prop in enumerate(property_list):
            # get name and cardinality.
            # GUI-order is equal to order in the sheet.
            property_ = {
                "propname": ":" + prop[0],
                "cardinality": str(prop[1]),
                "gui_order": i + 1
            }
            cards.append(property_)

        # build resource dict
        resource = {
            "name": name,
            "labels": labels,
            "super": sup,
            "cardinalities": cards
        }
        # append to resources list
        resources.append(resource)

    if validate_resources_with_schema(json.loads(json.dumps(resources, indent=4))):
        # write final list to JSON file if list passed validation
        with open(file=outfile, mode='w+', encoding='utf-8') as file:
            file.write(prefix)
            json.dump(resources, file, indent=4)
            print('Resource file was created successfully and written to file:', outfile)
    else:
        print('Resource data is not valid according to schema.')



import json
import re

from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union, Type
from rfc3987 import parse
from pprint import pprint

from .group import Group
from .helpers import IriTest, Actions, BaseError, Cardinality
from .langstring import LangString
from .connection import Connection
from .permission import PermissionValue, PermissionsIterator, Permissions

from .listnode import ListNode


@strict
class KnoraStandoffXml:
    """Used to handle XML strings for standoff markup"""

    __iriregexp = re.compile(r'IRI:[^:]*:IRI')
    __xmlstr: str

    def __init__(self, xmlstr: str) -> str:
        self.__xmlstr = str(xmlstr)

    def __str__(self) -> str:
        return self.__xmlstr

    def getXml(self) -> str:
        return self.__xmlstr

    def findall(self) -> Union[List[str], None]:
        return self.__iriregexp.findall(self.__xmlstr)

    def replace(self, fromStr: str, toStr: str) -> None:
        self.__xmlstr = self.__xmlstr.replace(fromStr, toStr)


@strict
class Value:
    _iri: Union[str, None]
    _comment: Union[str, None]
    _permissions: Union[Permissions, None]
    _upermission: Union[PermissionValue, None]
    _ark_url: Union[str, None]
    _vark_url: Union[str, None]

    def __init__(self,
                 iri: Optional[str] = None,
                 groups: Optional[Group] = None,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        self._iri = iri
        self._comment = comment
        self._permissions = permissions
        self._upermission = upermission
        self._ark_url = ark_url
        self._vark_url = vark_url

    def __str__(self):
        if self._iri:
            tmpstr = "(iri: " + self._iri
        else:
            tmpstr = ('(iri: -')
        if self._permissions:
            tmpstr = ", permissions: " + str(self._permissions)
        if self._comment:
            tmpstr += ", comment: " + self._comment
        tmpstr += ")"
        return tmpstr


    @property
    def iri(self) -> str:
        return self.iri

    @property
    def ark_url(self) -> str:
        return self._ark_url

    @property
    def vark_url(self) -> str:
        return self._vark_url

    @property
    def permissions(self) -> str:
        return self._permissions

    @property
    def upermission(self) -> str:
        return self._upermission

    @property
    def comment(self):
        return self._comment

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = {}
        if action == Actions.Create:
            if self._permissions is not None:
                tmp["knora-api:hasPermissions"] = self.permissions.toJsonLdObj()

            if self._comment is not None:
                tmp["knora-api:valueHasComment"] = str(self._comment)
        else:
            pass
        return tmp

    @staticmethod
    def get_typed_value(key: str, jsonld_obj: Any) -> Union[str, float]:
        try:
            tmp = jsonld_obj[key]
            if tmp.get("@type") == "xsd:decimal":
                result = float(tmp["@value"])
            elif tmp.get("@type") == "xsd:integer":
                result = int(tmp["@value"])
            elif tmp.get("@type") == "xsd:boolean":
                result = bool(tmp["@value"])
            elif tmp.get("@type") == "xsd:anyURI":
                result = str(tmp["@value"])
            elif tmp.get("@type") == "xsd:dateTimeStamp":
                result = str(tmp["@value"])
            elif tmp.get("@id") is not None:
                result = tmp["@id"]
            else:
                raise BaseError("Invalid data type in JSON-LD: \"{}\"!".format(tmp["@type"]))
            return result
        except KeyError as kerr:
            raise BaseError("Error in JSON-LD returned!")

    @staticmethod
    def getFromJsonLd(jsonld_obj) -> Dict[str, Union[str, float]]:

        return {
            'iri': jsonld_obj.get("@id"),
            'comment': jsonld_obj.get("knora-api:valueHasComment"),
            'ark_url': Value.get_typed_value("knora-api:arkUrl", jsonld_obj),
            'vark_url': Value.get_typed_value("knora-api:versionArkUrl", jsonld_obj),
            'permissions': Permissions.fromString(jsonld_obj.get("knora-api:hasPermissions")),
            'upermission': PermissionValue[jsonld_obj.get("knora-api:userHasPermission", jsonld_obj)]
        }


@strict
class TextValue(Value):
    _value: Union[str, KnoraStandoffXml]
    _mapping: str

    def __init__(self,
                 value: Union[str, KnoraStandoffXml],
                 mapping: Optional[str] = None,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        self._value = value
        self._mapping = mapping
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        return self._value

    @property
    def mapping(self) -> str:
        return self._mapping

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)

        if jsonld_obj.get("knora-api:textValueAsXml") is not None:
            tmp['mapping'] = jsonld_obj.get("knora-api:textValueHasMapping")
            tmp['value'] = jsonld_obj.get("knora-api:textValueAsXml")
        else:
            tmp['mapping'] = None
            tmp['value'] = jsonld_obj.get("knora-api:valueAsString")
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = 'knora-api:TextValue'
            if isinstance(self._value, KnoraStandoffXml):
                tmp['knora-api:textValueAsXml'] = self._value
                tmp['knora-api:textValueHasMapping'] = {
                    '@id': 'http://rdfh.ch/standoff/mappings/StandardMapping' if self._mapping is None else self._mapping
                }
            else:
                tmp['knora-api:valueAsString'] = str(self._value)
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return str(self._value)


@strict
class ColorValue(Value):
    _value: str

    def __init__(self,
                 value: str,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        #
        # a color value as used in HTML (e.g. "#aaccff"
        #
        m = re.match('^#(?:[0-9a-fA-F]{3}){1,2}$', str(value))
        if not m:
            raise BaseError("Invalid ColorValue format! " + str(value))
        self._value = str(value)
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = jsonld_obj.get("knora-api:colorValueAsColor")
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:ColorValue"
            tmp['knora-api:colorValueAsColor'] = self._value
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return self._value + ' ' + super().__str__()


@strict
class DateValue(Value):
    _calendar: str
    _e1: str
    _y1: int
    _m1: int
    _d1: int
    _e2: str
    _y2: int
    _m2: int
    _d2: int

    def __init__(self,
                 value: str,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        #
        # A knora date value
        #
        m = re.match(
            '(GREGORIAN:|JULIAN:)?(CE:|BCE:)?(\\d{4})?(-\\d{1,2})?(-\\d{1,2})?(:CE|:BCE)?(:\\d{4})?(-\\d{1,2})?(-\\d{1,2})?',
            str(value))
        if not m:
            raise BaseError("Invalid date format: \"{}\"!".format(str(value)))
        dp = m.groups()
        self._calendar = 'GREGORIAN' if dp[0] is None else dp[0].strip('-: ')
        self._e1 = 'CE' if dp[1] is None else dp[1].strip('-: ')
        self._y1 = None if dp[2] is None else int(dp[2].strip('-: '))
        self._m1 = None if dp[3] is None else int(dp[3].strip('-: '))
        self._d1 = None if dp[4] is None else int(dp[4].strip('-: '))
        self._e2 = 'CE' if dp[5] is None else dp[5].strip('-: ')
        self._y2 = None if dp[6] is None else int(dp[6].strip('-: '))
        self._m2 = None if dp[7] is None else int(dp[7].strip('-: '))
        self._d2 = None if dp[8] is None else int(dp[8].strip('-: '))
        if self._y1 is None:
            raise BaseError("Invalid date format! " + str(value))

        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        datestr = ''
        if self._calendar:
            datestr += self._calendar + ':'
        if self._e1:
            datestr += self._e1 + ':'
        if self._y1:
            datestr += str(self._y1) + '-'
        if self._m1:
            datestr += str(self._m1) + '-'
        if self._d1:
            datestr += str(self._d1)
        if self._e2:
            datestr += ':' + str(self._e2)
        if self._y2:
            datestr += ':' + str(self._y2)
        if self._m2:
            datestr += '-' + str(self._m2)
        if self._d2:
            datestr += '-' + str(self._d2)
        return datestr

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)

        datestr = ""
        datestr += jsonld_obj.get("knora-api:dateValueHasCalendar") + ":" \
            if jsonld_obj.get("knora-api:dateValueHasCalendar") is not None else ""
        datestr += jsonld_obj.get("knora-api:dateValueHasStartEra") + ":" \
            if jsonld_obj.get("knora-api:dateValueHasStartEra") is not None else ""
        datestr += str(jsonld_obj.get("knora-api:dateValueHasStartYear")) \
            if jsonld_obj.get("knora-api:dateValueHasStartYear") is not None else ""
        datestr += "-" + str(jsonld_obj.get("knora-api:dateValueHasStartMonth")) \
            if jsonld_obj.get("knora-api:dateValueHasStartMonth") is not None else ""
        datestr += "-" + str(jsonld_obj.get("knora-api:dateValueHasStartDay")) \
            if jsonld_obj.get("knora-api:dateValueHasStartDay") is not None else ""
        datestr += ":" + jsonld_obj.get("knora-api:dateValueHasEndEra") \
            if jsonld_obj.get("knora-api:dateValueHasEndEra") is not None else ""
        datestr += ":" + str(jsonld_obj.get("knora-api:dateValueHasEndYear")) \
            if jsonld_obj.get("knora-api:dateValueHasEndYear") is not None else ""
        datestr += "-" + str(jsonld_obj.get("knora-api:dateValueHasEndMonth")) \
            if jsonld_obj.get("knora-api:dateValueHasEndMonth") is not None else ""
        datestr += "-" + str(jsonld_obj.get("knora-api:dateValueHasEndDay")) \
            if jsonld_obj.get("knora-api:dateValueHasEndDay") is not None else ""
        tmp['value'] = datestr
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:DateValue"
            tmp["knora-api:dateValueHasCalendar"] = self._calendar
            tmp["knora-api:dateValueHasStartEra"] = self._e1
            tmp["knora-api:dateValueHasStartYear"] = self._y1
            if self._m1 is not None:
                tmp["knora-api:dateValueHasStartMonth"] = self._m1
            if self._d1 is not None:
                tmp["knora-api:dateValueHasStartDay"] = self._d1
            tmp["knora-api:dateValueHasEndEra"] = self._e2
            if self._y2 is not None:
                tmp["knora-api:dateValueHasEndYear"] = self._y2
            else:
                tmp["knora-api:dateValueHasEndYear"] = self._y1
            if self._m2 is not None:
                tmp["knora-api:dateValueHasEndMonth"] = self._m2
            if self._d2 is not None:
                tmp["knora-api:dateValueHasEndDay"] = self._d2
        else:
            pass
        return tmp

    def __str__(self):
        datestr = ''
        if self._calendar:
            datestr += self._calendar + ':'
        if self._e1:
            datestr += self._e1 + ':'
        if self._y1:
            datestr += str(self._y1) + '-'
        if self._m1:
            datestr += str(self._m1) + '-'
        if self._d1:
            datestr += str(self._d1)
        if self._e2:
            datestr += ':' + str(self._e2)
        if self._y2:
            datestr += ':' + str(self._y2)
        if self._m2:
            datestr += '-' + str(self._m2)
        if self._d2:
            datestr += '-' + str(self._d2)
        return datestr + ' ' + super().__str__()


@strict
class DecimalValue(Value):
    _value: float

    def __init__(self,
                 value: Union[float, int, str],
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        self._value = value
        if isinstance(value, str):
            m = re.match('^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$', value)
            if m:
                self._value = float(value)
            else:
                raise BaseError("String does not represent decimal/float number! \"" + value + "\"")
        elif isinstance(value, float):
            self._value = value
        elif isinstance(value, int):
            self._value = float(value)
        else:
            raise BaseError("String does not represent decimal/float number! \"" + value + "\"")
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> float:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = Value.get_typed_value("knora-api:decimalValueAsDecimal", jsonld_obj)
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:DecimalValue"
            tmp['knora-api:decimalValueAsDecimal'] = {
                '@type': 'xsd:decimal',
                '@value': str(self._value)
            }
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return str(self._value) + ' ' + super().__str__()


@strict
class GeomValue(Value):
    _value: str

    def __init__(self,
                 value: str,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        self._value = str(value)
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = jsonld_obj.get('knora-api:geometryValueAsGeometry')
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:GeomValue"
            tmp['knora-api:geometryValueAsGeometry'] = self._value
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return self._value + ' ' + super().__str__()

@strict
class GeonameValue(Value):
    _value: str

    def __init__(self,
                 value: str,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        self._value = str(value)
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = jsonld_obj.get('knora-api:geonameValueAsGeonameCode')
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:GeonameValue"
            tmp['knora-api:geonameValueAsGeonameCode'] = self._value
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return self._value + ' ' + super().__str__()


@strict
class IntValue(Value):
    _value: int

    def __init__(self,
                 value: Union[int, str],
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        if isinstance(value, str):
            m = re.match('^[-+]?[0-9]+$', value)
            if m and m.span()[1] == len(str(value)):
                self._value = int(value)
            else:
                raise BaseError("String does not represent integer number! \"" + value + "\"")
        elif isinstance(value, int):
            self._value = value
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> int:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = jsonld_obj.get("knora-api:intValueAsInt")
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:IntValue"
            tmp['knora-api:intValueAsInt'] = self._value
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return str(self._value) + ' ' + super().__str__()

@strict
class BooleanValue(Value):
    _value: bool

    def __init__(self,
                 value: Union[bool, int, str],
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        if type(value) is bool:
            self._value = value

        elif type(value) is str:
            if value.upper() == 'TRUE':
                self._value = True
            elif value.upper() == 'FALSE':
                self._value = False
            else:
                raise BaseError("Invalid boolean format! " + str(value))
        elif type(value) is int:
            self._value = False if value == 0 else True
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> bool:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = jsonld_obj.get("knora-api:booleanValueAsBoolean")
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:BooleanValue"
            tmp['knora-api:booleanValueAsBoolean'] = self._value
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return str(self._value) + ' ' + super().__str__()


@strict
class UriValue(Value):
    _value: str

    def __init__(self,
                 value: str,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        m = re.match("^(http)s?://([\\w\\.\\-~]+)?(:\\d{,6})?(/[\\w\\-~]+)*(#[\\w\\-~]*)?", str(value))
        if m:
            self._value = str(value)
        else:
            raise BaseError("Invalid IRI/URI! \"" + value + "\"")
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = Value.get_typed_value("knora-api:uriValueAsUri", jsonld_obj)
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:UriValue"
            tmp['knora-api:uriValueAsUri'] = {
                "@type": "xsd:anyURI",
                "@value": self._value
            }
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return self._value + ' ' + super().__str__()


@strict
class TimeValue(Value):
    _value: str

    @property
    def value(self) -> str:
        return self._value

    def __init__(self,
                 value: str,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        m = re.match("^([+-])?(\\d{4}-[0-1]\\d-[0-3]\\dT[0-2]\\d:[0-5]\\d:[0-5]\\d)(.\\d+)?(Z|[+-][0-2]\\d:[0-5]\\d)$",
                     str(value))
        if m:
            self._value = str(value)
        else:
            raise BaseError("Invalid time value! \"" + value + "\"")
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = Value.get_typed_value("knora-api:timeValueAsTimeStamp", jsonld_obj)
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:TimeValue"
            tmp['knora-api:timeValueAsTimeStamp'] = {
                "@type": "xsd:dateTimeStamp",
                "@value": self._value
            }
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return self._value + ' ' + super().__str__()


@strict
class IntervalValue(Value):
    _iv_start: str
    _iv_end: str

    def __init__(self,
                 value: Optional[str] = None,
                 iv_start: Optional[float] = None,
                 iv_end: Optional[float] = None,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        if value is not None:
            startval, endval = value.split(':')
            self._iv_start = float(startval)
            self._iv_end = float(endval)
        else:
            if iv_start is None or iv_end is None:
                raise BaseError("\"value\" or \"iv_start\" and \"iv_end\" must be given to constructor!")
            self._iv_start = iv_start
            self._iv_end = iv_end
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        return str(self._iv_start) + ':' + str(self._iv_end)

    @property
    def iv_start(self) -> float:
        return self._iv_start

    @property
    def iv_end(self) -> float:
        return self._iv_end

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        start = Value.get_typed_value("knora-api:intervalValueHasStart", jsonld_obj)
        end = Value.get_typed_value("knora-api:intervalValueHasEnd", jsonld_obj)
        tmp['value'] = str(start) + ":" + str(end)
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:IntervalValue"
            tmp["knora-api:intervalValueHasStart"] = {
                "@type": "xsd:decimal",
                "@value": str(self._iv_start)
            }
            tmp["knora-api:intervalValueHasEnd"] = {
                "@type": "xsd:decimal",
                "@value": str(self._iv_end)
            }
        else:
            pass
        return tmp

    def __str__(self) -> str:
        interval = str(self._iv_start) + ':' + str(self._iv_end)
        return interval + ' ' + super().__str__()


@strict
class ListValue(Value):
    _value: str

    def __init__(self,
                 value: str,
                 lists: List[ListNode] = None,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):

        def find_listnode(nodes: List[ListNode], name: str) -> Union[str, None]:
            for node in nodes:
                if node.name == name:
                    return node.id
                else:
                    if node.children is not None:
                        node_id = find_listnode(node.children, name)
                        if node_id is not None:
                            return node_id
            return None

        if IriTest.test(str(value)):
            self._value = str(value)
        else:
            tmp = str(value).split(':')
            if len(tmp) > 1:
                if tmp[0]:
                    listname = tmp[0]
                    nodename = tmp[1]
                else:
                    raise BaseError("Invalid list node: \"" + str(value) + "\" !")
            else:
                raise BaseError("Invalid list node: \"" + str(value) + "\" !")
            if lists is None:
                raise BaseError("Lists from ResourceInstanceFactory must be provided!")
            node_iri = None
            for list in lists:
                if list.name == listname:
                    node_iri = find_listnode(list.children, nodename)
            if node_iri is not None:
                self._value = node_iri
            else:
                raise BaseError("Listnode \"{}\" not found".format(value))
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    def print(self, offset: int = 0):
        blank = ' '
        print(f'{blank:>{offset}}{self._value}')

    @property
    def value(self) -> str:
        return self._value

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        tmp['value'] = Value.get_typed_value("knora-api:listValueAsListNode", jsonld_obj)
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:ListValue"
            tmp['knora-api:listValueAsListNode'] = {
                '@id': self._value
            }
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return self._value + ' ' + super().__str__()


@strict
class LinkValue(Value):
    _value: str
    _restype: str
    _reslabel: str

    def __init__(self,
                 value: str,
                 restype: Optional[str] = None,
                 reslabel: Optional[str] = None,
                 comment: Optional[LangString] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 iri: Optional[str] = None,
                 ark_url: Optional[str] = None,
                 vark_url: Optional[str] = None):
        m = re.match("^(http)s?://([\\w\\.\\-~]+)?(:\\d{,6})?(/[\\w\\-~]+)*(#[\\w\\-~]*)?", str(value))
        if m:
            self._value = str(value)
        else:
            raise BaseError("Target of link is not a IRI: \"" + str(value) + "\"")
        self._restype = restype
        self._reslabel = reslabel
        super().__init__(iri=iri,
                         comment=comment,
                         permissions=permissions,
                         upermission=upermission,
                         ark_url=ark_url,
                         vark_url=vark_url)

    @property
    def value(self) -> str:
        return self._value

    @property
    def restype(self) -> str:
        return self._restype

    @property
    def reslabel(self) -> str:
        return self._reslabel

    @classmethod
    def fromJsonLdObj(cls, jsonld_obj: Any) -> Dict[str, Any]:
        tmp = Value.getFromJsonLd(jsonld_obj)
        linked_resource = jsonld_obj.get("knora-api:linkValueHasTarget")
        if linked_resource is not None:
            tmp['value'] = linked_resource['@id']
            tmp['restype'] = linked_resource["@type"]
            tmp['reslabel'] = linked_resource["rdfs:label"]
        return cls(**tmp)

    def toJsonLdObj(self, action: Actions) -> Dict[str, Any]:
        tmp = super().toJsonLdObj(action)
        if action == Actions.Create:
            tmp['@type'] = "knora-api:LinkValue"
            tmp['knora-api:linkValueHasTargetIri'] = {
                '@id': self._value
            }
        else:
            pass
        return tmp

    def __str__(self) -> str:
        return self._value + ' ' + super().__str__()


def fromJsonLdObj(jsonld_obj: str) -> Value:
    switcher = {
        'knora-api:TextValue': TextValue,
        'knora-api:ColorValue': ColorValue,
        'knora-api:DateValue': DateValue,
        'knora-api:DecimalValue': DecimalValue,
        'knora-api:GeomValue': GeomValue,
        'knora-api:GeonameValue': GeonameValue,
        'knora-api:IntValue': IntValue,
        'knora-api:BooleanValue': BooleanValue,
        'knora-api:UriValue': UriValue,
        'knora-api:TimeValue': TimeValue,
        'knora-api:IntervalValue': IntervalValue,
        'knora-api:ListValue': ListValue,
        'knora-api:LinkValue': LinkValue,
    }
    return switcher[jsonld_obj.get('@type')].fromJsonLdObj(jsonld_obj)


def make_value(value: Value,
               comment: Optional[str] = None,
               permissions: Optional[Permissions] = None):
    res = {}
    res['value'] = value
    if comment:
        res['comment'] = comment
    if permissions:
        res['permissions'] = permissions
    return res



import json
import re
from dataclasses import dataclass
from enum import Enum, unique
from urllib.parse import quote_plus

from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union, Type
from copy import deepcopy

from .helpers import OntoInfo, Actions, BaseError, Cardinality, Context
from .connection import Connection
from .model import Model
from .project import Project
from .listnode import ListNode
from .ontology import Ontology
from .resourceclass import ResourceClass, HasProperty
from .permission import PermissionValue, PermissionsIterator, Permissions
from .value import KnoraStandoffXml, Value, TextValue, ColorValue, DateValue, DecimalValue, GeomValue, GeonameValue, \
    IntValue, BooleanValue, UriValue, TimeValue, IntervalValue, ListValue, LinkValue, fromJsonLdObj


from pprint import pprint


class KnoraStandoffXmlEncoder(json.JSONEncoder):
    """Classes used as wrapper for knora standoff-XML"""
    def default(self, obj) -> str:
        if isinstance(obj, KnoraStandoffXml):
            return '<?xml version="1.0" encoding="UTF-8"?>\n<text>' + obj.getXml() + '</text>'
        elif isinstance(obj, OntoInfo):
            return obj.iri + "#" if obj.hashtag else ""
        return json.JSONEncoder.default(self, obj)

@dataclass
class Propinfo:
    valtype: Type
    cardinality: Cardinality
    gui_order: int
    attributes: Optional[str] = None


@strict
class ResourceInstance(Model):
    baseclasses_with_bitstream: Set[str] = {
        'StillImageRepresentation',
        'AudioRepresentation',
        'DocumentRepresentation',
        'MovingImageRepresentation',
        'DDDRepresentation',
        'TextRepresentation'}
    knora_properties: Set[str] = {
        "knora-api:isPartOf",
        "knora-api:seqnum",
    }
    _iri: Union[str, None]
    _ark: Union[str, None]
    _vark: Union[str, None]
    _label: Union[str, None]
    _permissions: Union[Permissions, None]
    _upermission: Union[PermissionValue, None]
    _bitstream: Union[str, None]
    _values: Union[Dict[Value, List[Value]], None]

    def __init__(self,
                 con: Connection,
                 iri: Optional[str] = None,
                 ark: Optional[str] = None,
                 vark: Optional[str] = None,
                 label: Optional[str] = None,
                 permissions: Optional[Permissions] = None,
                 upermission: Optional[PermissionValue] = None,
                 bitstream: Optional[str] = None,
                 values: Optional[Dict[str, Union[str, List[str], Dict[str, str], List[Dict[str,str]], Value, List[Value]]]] = None):
        super().__init__(con)
        self._iri = iri
        self._label = label
        self._ark = ark
        self._vark = vark
        self._permissions = permissions
        self._upermission = upermission

        if self.baseclass in self.baseclasses_with_bitstream and bitstream is None:
            raise BaseError("The baseclass \"{}\" requires a bitstream value!".format(self.baseclass))
        if self.baseclass not in self.baseclasses_with_bitstream and bitstream is not None:
            raise BaseError("The baseclass \"{}\" does not allow a bitstream value!".format(self.baseclass))
        if self.baseclass in self.baseclasses_with_bitstream and bitstream is not None:
            self._bitstream = bitstream
        else:
            self._bitstream = None

        self._values = {}
        if values:
            self._values = {}
            for propname, propinfo in self.properties.items():
                # if propinfo.valtype is LinkValue:
                vals = values.get(propname)
                if vals is not None:
                    valcnt: int = 0
                    if type(vals) is list:  # we do have several values for this properties
                        self._values[propname] = []
                        for val in vals:
                            if valcnt > 0 and (propinfo.cardinality == Cardinality.C_0_1 or propinfo.cardinality == Cardinality.C_1):
                                raise BaseError(f'Cardinality does not allow multiple values for "{propname}"!')
                            if type(val) is Value:
                                self._values[propname].append(val)
                            elif type(val) is dict:
                                if propinfo.valtype is ListValue:
                                    val['lists'] = self.lists
                                self._values[propname].append(propinfo.valtype(**val))
                            else:
                                if propinfo.valtype is ListValue:
                                    val = {'value': val, 'lists': self.list}
                                self._values[propname].append(propinfo.valtype(val))
                            valcnt = valcnt + 1
                    else:  # we do have only one value for this property
                        if type(vals) is Value:
                            self._values[propname] = vals
                        elif type(vals) is dict:
                            if propinfo.valtype is ListValue:
                                vals['lists'] = self.lists
                            self._values[propname] = propinfo.valtype(**vals)
                        else:
                            if propinfo.valtype is ListValue:
                                vals = {'value': val, 'lists': self.list}
                            self._values[propname] = propinfo.valtype(vals)
                else:
                    if propinfo.cardinality == Cardinality.C_1 or propinfo.cardinality == Cardinality.C_1_n:
                        raise BaseError("Cardinality does require at least one value for \"{}\"!".format(propname))
            for propname in values:
                if propname not in self.knora_properties and self.properties.get(propname) is None:
                    raise BaseError(f'Property "{propname}" is not part of data model!')

    def value(self, item):
        if self._values.get(item):
            val = self._values[item]
            if isinstance(val, list):
                tmp = [x.value for x in val]
                return tmp
            else:
                return val.value
        else:
            return None

    @property
    def label(self) -> str:
        return self._label

    @property
    def iri(self) -> str:
        return self._iri

    @property
    def ark(self) -> str:
        return self._ark

    @property
    def vark(self) -> str:
        return self._vark

    def clone(self) -> 'ResourceInstance':
        return deepcopy(self)

    def fromJsonLdObj(self, con: Connection, jsonld_obj: Any) -> 'ResourceInstance':
        newinstance = self.clone()
        newinstance._iri = jsonld_obj.get('@id')
        resclass = jsonld_obj.get('@type')
        context = Context(jsonld_obj.get('@context'))
        newinstance._label = jsonld_obj.get("rdfs:label")
        newinstance._ark = Value.get_typed_value("knora-api:arkUrl", jsonld_obj)
        newinstance._vark = Value.get_typed_value("knora-api:versionArkUrl", jsonld_obj)
        newinstance._permissions = Permissions.fromString(jsonld_obj.get("knora-api:hasPermissions"))
        newinstance._upermission = PermissionValue[jsonld_obj.get("knora-api:userHasPermission", jsonld_obj)]
        creation_date = Value.get_typed_value("knora-api:creationDate", jsonld_obj)
        user = Value.get_typed_value("knora-api:attachedToUser", jsonld_obj)
        project = Value.get_typed_value("knora-api:attachedToProject", jsonld_obj)
        to_be_ignored = [
            '@id', '@type', '@context', "rdfs:label", "knora-api:arkUrl", "knora-api:versionArkUrl",
            "knora-api:creationDate", "knora-api:attachedToUser", "knora-api:attachedToProject",
            "knora-api:hasPermissions", "knora-api:userHasPermission"
        ]
        if id is None:
            raise BaseError('Resource "id" is missing in JSON-LD from DSP-API')
        type = jsonld_obj.get('@type')
        newinstance._values: Dict[str, Union[Value, List[Value]]] = {}
        for key, obj in jsonld_obj.items():
            if key in to_be_ignored:
                continue
            try:
                if isinstance(obj, list):
                    newinstance._values[key] = []
                    for o in obj:
                        newinstance._values[key].append(fromJsonLdObj(o))
                else:
                    newinstance._values[key] = fromJsonLdObj(obj)
            except KeyError as kerr:
                raise BaseError("Invalid data in JSON-LD: \"{}\" has value class \"{}\"!".format(key, obj.get("@type")))
        return newinstance

    def toJsonLdObj(self, action: Actions) -> Any:
        tmp = {}
        if action == Actions.Create:
            tmp['@type'] = self.classname
            tmp["knora-api:attachedToProject"] = {
                "@id": self.project
            }
            tmp['rdfs:label'] = self._label
            if self._permissions:
                tmp["knora-api:hasPermissions"] = self._permissions.toJsonLdObj()
            if self._bitstream:
                if self.baseclass == 'StillImageRepresentation':
                    tmp["knora-api:hasStillImageFileValue"] = {
                        "@type": "knora-api:StillImageFileValue",
                        "knora-api:fileValueHasFilename": self._bitstream
                    }
                elif self.baseclass == 'DocumentRepresentation':
                    tmp["knora-api:hasDocumentFileValue"] = {
                        "@type": "knora-api:DocumentFileValue",
                        "knora-api:fileValueHasFilename": self._bitstream
                    }
                elif self.baseclass == 'TextRepresentation':
                    tmp["knora-api:hasTextFileValue"] = {
                        "@type": "knora-api:TextFileValue",
                        "knora-api:fileValueHasFilename": self._bitstream
                    }
                elif self.baseclass == 'AudioRepresentation':
                    tmp["knora-api:hasAudioFileValue"] = {
                        "@type": "knora-api:AudioFileValue",
                        "knora-api:fileValueHasFilename": self._bitstream
                    }
                else:
                    raise BaseError(f'Baseclass "{self.baseclass}" not yet supported!')
            for propname, valtype in self._values.items():
                if type(valtype) is list:
                    if type(valtype[0]) is LinkValue:
                        propname += 'Value'
                    tmp[propname] = []
                    for vt in valtype:
                        tmp[propname].append(vt.toJsonLdObj(action))
                    pass
                else:
                    if type(valtype) is LinkValue:
                        propname += 'Value'
                    tmp[propname] = valtype.toJsonLdObj(action)
            tmp['@context'] = self.context
        else:
            pass
        return tmp

    def create(self):
        jsonobj = self.toJsonLdObj(Actions.Create)
        jsondata = json.dumps(jsonobj, indent=4, separators=(',', ': '), cls=KnoraStandoffXmlEncoder)
        #print(jsondata)
        result = self._con.post('/v2/resources', jsondata)
        newinstance = self.clone()
        newinstance._iri = result['@id']
        newinstance._ark = result['knora-api:arkUrl']['@value']
        newinstance._vark = result['knora-api:versionArkUrl']['@value']
        return newinstance

    def read(self) -> 'ResourceInstance':
        result = self._con.get('/v2/resources/' + quote_plus(self._iri))
        return self.fromJsonLdObj(con=self._con, jsonld_obj=result)


    def update(self):
        pass

    def delete(self):
        pass

    def print(self):
        print('Iri:', self._iri)
        print('Ark:', self._ark)
        print('Vark:', self._vark)
        print('Label:', self._label)
        print('Permissions:', str(self._permissions))
        print('Userpermission:', str(self._upermission))
        for name, val in self._values.items():
            if isinstance(val, list):
                tmp = [str(x) for x in val]
                print(name, ':', " | ".join(tmp))
                pass
            else:
                print(name, ':', str(val))


# ToDo: special resourceclasses and properties of knora-api
#
# - knora-api:isPartOf -> Object: knora-api:Resource
# - knora-api:author -> Object: knora-api:User
# - knora-api:seqnum -> Object: knora-api:IntValue
# - knora-api:hasComment -> Object: knora-api:TextValue
#
# knora-api:Region: IS a resource
# - knora-api:hasGeometry
# - knora-api:isRegionOf
# - knora-api:hasColor
# - knora-api:hasComment

@strict
class ResourceInstanceFactory:
    _con: Connection
    _project: Project
    _lists = List[ListNode]
    _ontologies = Dict[str, Ontology]
    _ontoname2iri = Dict[str, str]
    _context: Context

    def __init__(self,
                 con: Connection,
                 projident: str):
        self._con = con
        if re.match("^[0-9aAbBcCdDeEfF]{4}$", projident):
            project = Project(con=self._con, shortcode=projident)
        elif re.match("^[\\w-]+$", projident):
            project = Project(con=self._con, shortname=projident)
        elif re.match("^(http)s?://([\\w\\.\\-~]+:?\\d{,4})(/[\\w\\-~]+)+$", projident):
            project = Project(con=self._con, shortname=projident)
        else:
            raise BaseError("Invalid project identification!")
        self._project = project.read()

        tmp = ListNode.getAllLists(con=self._con, project_iri=self._project.id)
        self._lists = []
        for rnode in tmp:
            self._lists.append(rnode.getAllNodes())

        tmp_ontologies = Ontology.getProjectOntologies(con, self._project.id)
        shared_project = Project(con=self._con, shortcode="0000").read()
        shared_ontologies = Ontology.getProjectOntologies(con, shared_project.id)
        tmp_ontologies.extend(shared_ontologies)
        self._ontoname2iri = {x.name: x.id for x in tmp_ontologies}

        ontology_ids = [x.id for x in tmp_ontologies]
        self._ontologies = {}
        self._properties = {}
        self._context = {}
        for onto in ontology_ids:
            oparts = onto.split("/")
            name = oparts[len(oparts) - 2]
            shortcode = oparts[len(oparts) - 3]
            self._ontologies[name] = Ontology.getOntologyFromServer(con=con, shortcode=shortcode, name=name)
            self._properties.update({name + ':' + x.name: x for x in self._ontologies[name].property_classes})
            self._context.update(self._ontologies[name].context)

    @property
    def lists(self) -> List[ListNode]:
        return self._lists

    def get_resclass_names(self) -> List[str]:
        resclass_names: List[str] = []
        for name, onto in self._ontologies.items():
            for resclass in onto.resource_classes:
                resclass_names.append(onto.context.get_prefixed_iri(resclass.id))
        return resclass_names

    def _get_baseclass(self, superclasses: List[str]) -> Union[str, None]:
        for sc in superclasses:
            ontoname, classname = sc.split(':')
            if ontoname == 'knora-api':
                return classname
            o = self._ontologies.get(ontoname)
            if o is None:
                continue
            gaga = [x for x in o.resource_classes if x.name == classname][0]
            return self._get_baseclass(gaga.superclasses)
        return None

    def get_resclass(self, prefixedresclass: str) -> Type:
        prefix, resclass_name = prefixedresclass.split(':')
        resclass = [x for x in self._ontologies[prefix].resource_classes if x.name == resclass_name][0]
        baseclass = self._get_baseclass(resclass.superclasses)
        props: Dict[str, Propinfo] = {}
        switcher = {
            'knora-api:TextValue': TextValue,
            'knora-api:ColorValue': ColorValue,
            'knora-api:DateValue': DateValue,
            'knora-api:DecimalValue': DecimalValue,
            'knora-api:GeomValue': GeomValue,
            'knora-api:GeonameValue': GeonameValue,
            'knora-api:IntValue': IntValue,
            'knora-api:BooleanValue': BooleanValue,
            'knora-api:UriValue': UriValue,
            'knora-api:TimeValue': TimeValue,
            'knora-api:IntervalValue': IntervalValue,
            'knora-api:ListValue': ListValue,
            'knora-api:LinkValue': LinkValue,
        }
        for propname, has_property in resclass.has_properties.items():
            if propname == "knora-api:isPartOf":
                valtype = LinkValue
                props[propname] = Propinfo(valtype=valtype,
                                           cardinality=has_property.cardinality,
                                           gui_order=has_property.gui_order)
            elif propname == "knora-api:seqnum":
                valtype = IntValue
                props[propname] = Propinfo(valtype=valtype,
                                           cardinality=has_property.cardinality,
                                           gui_order=has_property.gui_order)
            elif propname == "knora-api:hasComment":
                valtype = TextValue
                props[propname] = Propinfo(valtype=valtype,
                                           cardinality=has_property.cardinality,
                                           gui_order=has_property.gui_order)
            elif has_property.ptype == HasProperty.Ptype.other:
                valtype = switcher.get(self._properties[propname].object)
                if valtype == LinkValue:
                    continue  # we have the Link to the LinkValue which we do not use
                if valtype is None:
                    valtype = LinkValue
                    props[propname] = Propinfo(valtype=valtype,
                                               cardinality=has_property.cardinality,
                                               gui_order=has_property.gui_order,
                                               attributes=self._properties[propname].object)
                else:
                    props[propname] = Propinfo(valtype=valtype,
                                               cardinality=has_property.cardinality,
                                               gui_order=has_property.gui_order)
        return type(resclass_name, (ResourceInstance,), {'project': self._project.id,
                                                         'classname': prefixedresclass,
                                                         'baseclass': baseclass,
                                                         'context': self._context,
                                                         'properties': props,
                                                         'lists': self._lists})





import unittest
import pprint

from dsplib.models.connection import Connection
from dsplib.models.helpers import BaseError, Actions, LastModificationDate
from dsplib.models.langstring import Languages, LangStringParam, LangString
from dsplib.models.ontology import Ontology
from dsplib.models.propertyclass import PropertyClass


class TestPropertyClass(unittest.TestCase):
    project = "http://rdfh.ch/projects/0001"
    onto_name = 'propclass-test-a'
    onto_label = 'propclass_test_ontology'

    onto: Ontology
    last_modification_date: LastModificationDate
    con: Connection

    name = 'MyPropClassName'
    object = 'TextValue'
    label = LangString({Languages.DE: 'MyPropClassLabel'})
    comment = LangString({Languages.DE: 'This is a property class for testing'})


    def setUp(self) -> None:
        #
        # Connect to Knora
        #
        self.con = Connection('http://0.0.0.0:3333')
        self.con.login('root@example.com', 'test')

        ontos = Ontology.getAllOntologies(self.con)
        for onto in ontos:
            onto.print(True)
        #
        # Create a test ontology
        #
        self.onto = Ontology(
            con=self.con,
            project=self.project,
            name=self.onto_name,
            label=self.onto_label,
        ).create()
        self.assertIsNotNone(self.onto.id)
        self.last_modification_date = self.onto.lastModificationDate

    def tearDown(self):
        #
        # remove test ontology
        #
        result = self.onto.delete()
        self.assertIsNotNone(result)


    def test_PropertyClass_create(self):
        #
        # Create new property class
        #
        self.last_modification_date, propclass = PropertyClass(
            con=self.con,
            context=self.onto.context,
            name=self.name,
            ontology_id=self.onto.id,
            object=self.object,
            label=self.label,
            comment=self.comment
        ).create(self.last_modification_date)
        self.onto.lastModificationDate = self.last_modification_date
        self.assertIsNotNone(propclass.id)

        self.assertEqual(propclass.name, self.name)
        self.assertEqual(propclass.label['de'], self.label['de'])
        self.assertEqual(propclass.comment['de'], self.comment['de'])

        #
        # Again get ontology data
        #
        self.onto = self.onto.read()
        self.last_modification_date = self.onto.lastModificationDate
        self.last_modification_date = propclass.delete(self.last_modification_date)

        #
        # Again get ontology data
        #
        self.onto = self.onto.read()

    def test_PropertyClass_update(self):
        self.onto = self.onto.read()

        #
        # create test resource class
        #
        self.last_modification_date, propclass = PropertyClass(
            con=self.con,
            context=self.onto.context,
            name=self.name,
            ontology_id=self.onto.id,
            object=self.object,
            label=self.label,
            comment=self.comment
        ).create(self.last_modification_date)
        self.onto.lastModificationDate = self.last_modification_date
        self.assertIsNotNone(propclass.id)

        #
        # Modify the property class
        #
        propclass.addLabel('en', "This is english gaga")
        propclass.rmLabel('de')
        propclass.addComment('it', "Commentario italiano")
        self.last_modification_date, propclass = propclass.update(self.last_modification_date)
        self.onto.lastModificationDate = self.last_modification_date
        self.assertEqual(propclass.label['en'], "This is english gaga")
        self.assertEqual(propclass.comment['it'], "Commentario italiano")

        #
        # Now delete the resource class to clean up
        #
        self.last_modification_date = propclass.delete(self.last_modification_date)
        self.onto.lastModificationDate = self.last_modification_date


if __name__ == '__main__':
    unittest.main()



import re
from enum import Enum, unique
from typing import List, Dict, Optional, Union

from pystrict import strict


@unique
class PermissionValue(Enum):
    RV = 1
    V = 2
    M = 4
    D = 8
    CR = 16

    def __str__(self):
        tmp = {
            1: 'RV',
            2: 'V',
            4: 'M',
            8: 'D',
            16: 'CR'
        }
        return tmp[self.value]


@strict
class PermissionsIterator:
    _permissions: 'Permissions'
    _group: List[str]
    _index: int

    def __init__(self, permissions: 'Permissions'):
        self._permissions = permissions
        self._index = 0

    def __next__(self):
        if len(self._permissions.permissions) == 0 and self._index == 0:
            return None, None
        elif self._index < len(self._permissions.permissions):
            tmp = self._prefixes[self._index]
            self._index += 1
            return tmp, self._permissions.permissions[tmp]
        else:
            raise StopIteration


@strict
class Permissions:
    _permissions: Union[Dict[PermissionValue, List[str]], None]

    def __init__(self,
                 permissions: Optional[Dict[PermissionValue, List[str]]] = None):
        if permissions is None:
            self._permissions = {}
        else:
            self._permissions = permissions

    def __getitem__(self, key: PermissionValue) -> Union[List[str], None]:
        return self._permissions.get(key)

    def __setitem__(self, key: PermissionValue, value: List[str]) -> None:
        self._permissions[key] = value

    def __delitem__(self, key: PermissionValue) -> None:
        del self._permissions[key]

    def __missing__(self, key: PermissionValue) -> None:
        return None

    def __iter__(self) -> PermissionsIterator:
        return PermissionsIterator(self)

    def __contains__(self, key: PermissionValue) -> bool:
        return key in self._permissions

    def __str__(self):
        tmpstr = ''
        for permission, groups in self._permissions.items():
            if tmpstr:
                tmpstr += '|'
            tmpstr += str(permission) + ' ' + ",".join(groups)
        return tmpstr

    def add(self, key: PermissionValue, val: str):
        if self._permissions.get(key) is None:
            self._permissions[key] = [val]
        else:
            self._permissions[key].append(val)

    def toJsonLdObj(self):
        tmpstr = ''
        for permission, groups in self._permissions.items():
            if tmpstr:
                tmpstr += '|'
            tmpstr += str(permission) + ' ' + ",".join(groups)
        return tmpstr

    @classmethod
    def fromString(cls, permstr: str):
        tmpstr = permstr.split('|')
        permissions: Dict[PermissionValue, List[str]] = {}
        for s in tmpstr:
            key, *vals = re.split("[\\s,]+", s)
            permissions[PermissionValue[key]] = vals
        return cls(permissions)

    @property
    def permissions(self) -> Union[Dict[PermissionValue, List[str]], None]:
        return self._permissions



import json
import os

import jsonschema
from openpyxl import load_workbook


def validate_properties_with_schema(json_file: str) -> bool:
    """
        This function checks if the json properties are valid according to the schema.

        Args:
            json_file: the json with the properties to be validated

        Returns:
            True if the data passed validation, False otherwise

        """
    current_dir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(current_dir, 'knora-schema-properties-only.json')) as schema:
        properties_schema = json.load(schema)

    try:
        jsonschema.validate(instance=json_file, schema=properties_schema)
    except jsonschema.exceptions.ValidationError as err:
        print(err)
        return False
    print('Properties data passed schema validation.')
    return True


def properties_excel2json(excelfile: str, outfile: str):
    """
        Converts properties described in an Excel file into a properties section which can be integrated into a DSP ontology

        Args:
            excelfile: path to the Excel file containing the properties
            outfile: path to the output JSON file containing the properties section for the ontology

        Returns:
            None
    """
    # load file
    wb = load_workbook(filename=excelfile, read_only=True)
    sheet = wb.worksheets[0]
    props = [row_to_prop(row) for row in sheet.iter_rows(min_row=2, values_only=True, max_col=9)]

    prefix = '"properties":'

    if validate_properties_with_schema(json.loads(json.dumps(props, indent=4))):
        # write final list to JSON file if list passed validation
        with open(file=outfile, mode='w+', encoding='utf-8') as file:
            file.write(prefix)
            json.dump(props, file, indent=4)
            print('Properties file was created successfully and written to file:', outfile)
    else:
        print('Properties data is not valid according to schema.')

    return props


def row_to_prop(row):
    """
    Parses the row of an Excel sheet and makes a property from it

    Args:
        row: the row of an Excel sheet

    Returns:
        prop (JSON): the property in JSON format
    """
    name, super_, object_, en, de, fr, it, gui_element, hlist = row
    labels = {}
    if en:
        labels['en'] = en
    if de:
        labels['de'] = de
    if fr:
        labels['fr'] = fr
    if it:
        labels['it'] = it
    if not labels:
        raise Exception(f"No label given in any of the four languages: {name}")
    prop = {
        'name': name,
        'super': [super_],
        'object': object_,
        'labels': labels,
        'gui_element': gui_element
    }
    if hlist:
        prop['gui_attributes'] = {'hlist': hlist}
    return prop



import json
import copy
from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union
from urllib.parse import quote_plus

from .connection import Connection
from .helpers import Actions, BaseError, Context, LastModificationDate, OntoInfo, WithId
from .model import Model
from .project import Project
from .propertyclass import PropertyClass
from .resourceclass import ResourceClass

class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        elif isinstance(obj, Context):
            return obj.toJsonObj()
        elif isinstance(obj, OntoInfo):
            return {"iri": obj.iri, "hashtag": obj.hashtag}
        return json.JSONEncoder.default(self, obj)


"""
This model implements the handling of ontologies. It is to note that ResourceClasses, PropertyClasses
as well as the assignment of PropertyCLasses to the ResourceClasses (with a given cardinality)
is handeld in "cooperation" with the propertyclass.py (PropertyClass) and resourceclass.py (ResourceClass
and HasProperty) modules.

_Note_: All modifications to an ontology

CREATE:
    * Instantiate a new object of the Ontology class with all required parameters
    * Call the ``create``-method on the instance to create the ontology withing the backend

READ:
    * Instantiate a new object with ``id``(IRI of ontology) given
    * Call the ``read``-method on the instance. Reading the ontology also reads all
      associated PropertyClasses and ResourceClasses as well as the assignments.
    * Access the information that has been provided to the instance

UPDATE:
    * You need an instance of an existing Ontology by reading an instance
    * Change the attributes by assigning the new values
    * Call the ``update```method on the instance

DELETE
    * Instantiate a new objects with ``id``(IRI of group) given, or use any instance that has the id set,
      that is, that You've read before
    * Call the ``delete``-method on the instance

"""
@strict
class Ontology(Model):
    _id: str
    _project: str
    _name: str
    _label: str
    _comment: str
    _lastModificationDate: LastModificationDate
    _resource_classes: List[ResourceClass]
    _property_classes: List[PropertyClass]
    _context: Context
    _skiplist: List[str]

    def __init__(self,
                 con:  Connection,
                 id: Optional[str] = None,
                 project: Optional[Union[str, Project]] = None,
                 name: Optional[str] = None,
                 label: Optional[str] = None,
                 comment: Optional[str] = None,
                 lastModificationDate: Optional[Union[str, LastModificationDate]] = None,
                 resource_classes: List[ResourceClass] = [],
                 property_classes: List[PropertyClass] = [],
                 context: Context = None):
        super().__init__(con)
        self._id = id
        if isinstance(project, Project):
            self._project = project.id
        else:
            self._project = project
        self._name = name
        self._label = label
        self._comment = comment
        if lastModificationDate is None:
            self._lastModificationDate = None
        elif isinstance(lastModificationDate, LastModificationDate):
            self._lastModificationDate = lastModificationDate
        else:
            self._lastModificationDate = LastModificationDate(lastModificationDate)
        self._resource_classes = resource_classes
        self._property_classes = property_classes
        self._context = context if context is not None else Context()
        self._skiplist = []

    @property
    def id(self) -> str:
        return self._id

    @id.setter
    def id(self, value: str):
        raise BaseError('Ontology id cannot be modified!')

    @property
    def project(self) -> str:
        return self._project

    @project.setter
    def project(self, value: str):
        raise BaseError("An ontology's project cannot be modified!")

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        raise BaseError("An ontology's name cannot be modified!")

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value: str):
        self._label = str(value)
        self._changed.add('label')

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, value: str):
        self._comment = str(value)
        self._changed.add('comment')

    @property
    def lastModificationDate(self) -> LastModificationDate:
        return self._lastModificationDate

    @lastModificationDate.setter
    def lastModificationDate(self, value: Union[str, LastModificationDate]):
        self._lastModificationDate = LastModificationDate(value)

    @property
    def resource_classes(self) -> List[ResourceClass]:
        return self._resource_classes

    @resource_classes.setter
    def resource_classes(self, value: List[ResourceClass]) -> None:
        self._resource_classes = value

    def addResourceClass(self, resourceclass: ResourceClass, create: bool = False) -> Tuple[int, ResourceClass]:
        if create:
            print('Calling resourceclass.create in Ontology.addResourceClass')
            lmd, resourceclass = resourceclass.create(self._lastModificationDate)
            self._lastModificationDate = lmd
        self._resource_classes.append(resourceclass)
        index = len(self._resource_classes) - 1
        return index, resourceclass

    def updateResourceClass(self, index: int, resourceclass: ResourceClass) -> ResourceClass:
        lmd, resourceclass = resourceclass.update(self._lastModificationDate)
        self._lastModificationDate = lmd
        self._resource_classes[index] = resourceclass
        return resourceclass

    def removeResourceClass(self, index: int, erase: bool = False) -> None:
        if erase:
            lmd = self._resource_classes[index].delete(self._lastModificationDate)
            self._lastModificationDate = lmd
        del self._resource_classes[index]


    @property
    def property_classes(self) -> List[PropertyClass]:
        return self._property_classes

    @property_classes.setter
    def property_classes(self, value: List[PropertyClass]):
        self._property_classes = value

    def addPropertyClass(self, propclass: PropertyClass, create: bool = False) -> Tuple[int, ResourceClass]:
        if create:
            lmd, resourceclass = propclass.create(self._lastModificationDate)
            self._lastModificationDate = lmd
        self._property_classes.append(resourceclass)
        index = len(self._property_classes) - 1
        return index, propclass

    def updatePropertyClass(self, index: int, propclass: PropertyClass) -> PropertyClass:
        lmd, propclass = propclass.update(self._lastModificationDate)
        self._lastModificationDate = lmd
        self._property_classes[index] = propclass
        return propclass

    def removePropertyClass(self, index: int, erase: bool = False) -> None:
        if erase:
            lmd = self._property_classes[index].delete(self._lastModificationDate)
            self._lastModificationDate = lmd
        del self._property_classes[index]

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value: Context):
        raise BaseError('"Context" cannot be set!')

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any) -> 'Ontology':
        #
        # First let's get the ID (IRI) of the ontology
        #
        id = json_obj.get('@id')
        if id is None:
            raise BaseError('Ontology id is missing')

        #
        # evaluate the JSON-LD context to get the proper prefixes
        #
        context = Context(json_obj.get('@context'))
        onto_name = id.split('/')[-2]
        context.add_context(onto_name, id + '#')
        rdf = context.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        xsd = context.prefix_from_iri("http://www.w3.org/2001/XMLSchema#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui = context.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")
        this_onto = context.prefix_from_iri(id + "#")

        label = json_obj.get(rdfs + ':label')
        if label is None:
            raise BaseError('Ontology label is missing')
        comment = json_obj.get(rdfs + ':comment')
        if json_obj.get(knora_api + ':attachedToProject') is None:
            raise BaseError('Ontology not attached to a project')
        if json_obj[knora_api + ':attachedToProject'].get('@id') is None:
            raise BaseError('Ontology not attached to a project')
        project = json_obj[knora_api + ':attachedToProject']['@id']
        tmp = json_obj.get(knora_api + ':lastModificationDate')
        if tmp is not None:
            last_modification_date = LastModificationDate(json_obj.get(knora_api + ':lastModificationDate'))
        else:
            last_modification_date = None
        resource_classes = None
        property_classes = None
        if json_obj.get('@graph') is not None:
            resclasses_obj = list(filter(lambda a: a.get(knora_api + ':isResourceClass') is not None, json_obj.get('@graph')))
            resource_classes = list(map(lambda a: ResourceClass.fromJsonObj(con=con,
                                                                            context=context,
                                                                            json_obj=a), resclasses_obj))
            standoffclasses_obj = list(filter(lambda a: a.get(knora_api + ':isStandoffClass') is not None, json_obj.get('@graph')))
            # ToDo: parse standoff classes

            properties_obj = list(filter(lambda a: a.get(knora_api + ':isResourceProperty') is not None, json_obj.get('@graph')))
            #property_classes = list(map(lambda a: PropertyClass.fromJsonObj(con=con,
            #                                                                context=context,
            #                                                                json_obj=a), properties_obj))
            property_classes = [PropertyClass.fromJsonObj(con=con, context=context, json_obj=a) for a in properties_obj
                                if WithId(a.get(knora_api + ':objectType')).str() != "knora-api:LinkValue"]
        return cls(con=con,
                   id=id,
                   label=label,
                   project=project,
                   name=this_onto,  # TODO: corresponds the prefix always to the ontology name?
                   comment=comment,
                   lastModificationDate=last_modification_date,
                   resource_classes=resource_classes,
                   property_classes=property_classes,
                   context=context)

    @classmethod
    def __oneOntologiesFromJsonObj(cls, con: Connection, json_obj: Any, context: Context) -> 'Ontology':
        rdf = context.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        xsd = context.prefix_from_iri("http://www.w3.org/2001/XMLSchema#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui = context.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")
        if json_obj.get('@type') != owl + ':Ontology':
            return None
        id = json_obj.get('@id')
        if id is None:
            raise BaseError('Ontology id is missing')
        if json_obj.get(knora_api + ':attachedToProject') is None:
            raise BaseError('Ontology not attached to a project (1)')
        if json_obj[knora_api + ':attachedToProject'].get('@id') is None:
            raise BaseError('Ontology not attached to a project (2)')
        project = json_obj[knora_api + ':attachedToProject']['@id']
        tmp = json_obj.get(knora_api + ':lastModificationDate')
        if tmp is not None:
            last_modification_date = LastModificationDate(json_obj.get(knora_api + ':lastModificationDate'))
        else:
            last_modification_date = None
        label = json_obj.get(rdfs + ':label')
        if label is None:
            raise BaseError('Ontology label is missing')
        comment = json_obj.get(rdfs + ':comment')
        this_onto = id.split('/')[-2]
        context2 = copy.deepcopy(context)
        context2.add_context(this_onto, id)
        return cls(con=con,
                   id=id,
                   label=label,
                   project=project,
                   name=this_onto,
                   comment=comment,
                   lastModificationDate=last_modification_date,
                   context=context2)

    @classmethod
    def allOntologiesFromJsonObj(cls, con: Connection, json_obj: Any) -> List['Ontology']:
        context = Context(json_obj.get('@context'))
        rdf = context.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        xsd = context.prefix_from_iri("http://www.w3.org/2001/XMLSchema#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui = context.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")
        ontos: List['Ontology'] = []
        if json_obj.get('@graph') is not None:
            for o in json_obj['@graph']:
                ontos.append(Ontology.__oneOntologiesFromJsonObj(con, o, context))
        else:
            onto = Ontology.__oneOntologiesFromJsonObj(con, json_obj, context)
            if onto is not None:
                ontos.append(onto)
        return ontos

    def toJsonObj(self, action: Actions) -> Any:
        rdf = self._context.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs = self._context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = self._context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        xsd = self._context.prefix_from_iri("http://www.w3.org/2001/XMLSchema#")
        knora_api = self._context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui = self._context.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")
        # this_onto = self._context.prefixFromIri(self._id + "#")
        tmp = {}
        if action == Actions.Create:
            if self._name is None:
                raise BaseError('There must be a valid name given!')
            if self._label is None:
                raise BaseError('There must be a valid label given!')
            if self._project is None:
                raise BaseError('There must be a valid project given!')
            tmp = {
                knora_api + ":ontologyName": self._name,
                knora_api + ":attachedToProject": {
                    "@id": self._project
                },
                rdfs + ":label": self._label,
                "@context": self._context.toJsonObj()
            }
            if self._comment is not None:
                tmp[rdfs + ":comment"] = self._comment
        elif action == Actions.Update:
            if self._lastModificationDate is None:
                raise BaseError("'last_modification_date' must be in ontology!")
            tmp = {
                '@id': self._id,
                rdfs + ':label': self._label,
                knora_api + ':lastModificationDate': self._lastModificationDate.toJsonObj(),
                "@context": self._context.toJsonObj()
            }
            if self._label is not None and 'label' in self._changed:
                tmp[rdfs + ':label'] = self._label
            if self._comment is not None and 'comment' in self._changed:
                tmp[rdfs + ':comment'] = self._comment
        return tmp

    def create(self, dumpjson: Optional[str] = None) -> 'Ontology':
        jsonobj = self.toJsonObj(Actions.Create)
        jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=4)
        result = self._con.post('/v2/ontologies', jsondata)
        return Ontology.fromJsonObj(self._con, result)

    def update(self) -> 'Ontology':
        jsonobj = self.toJsonObj(Actions.Update)
        jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=4)
        result = self._con.put('/v2/ontologies/metadata', jsondata, 'application/ld+json')
        return Ontology.fromJsonObj(self._con, result)

    def read(self) -> 'Ontology':
        result = self._con.get('/v2/ontologies/allentities/' + quote_plus(self._id) + '?allLanguages=true')
        return Ontology.fromJsonObj(self._con, result)

    def delete(self) -> Optional[str]:
        result = self._con.delete('/v2/ontologies/' + quote_plus(self._id),
                                   params={'lastModificationDate': str(self._lastModificationDate)})
        return result.get('knora-api:result')

    @staticmethod
    def getAllOntologies(con: Connection) -> List['Ontology']:
        result = con.get('/v2/ontologies/metadata/')
        return Ontology.allOntologiesFromJsonObj(con, result)

    @staticmethod
    def getProjectOntologies(con: Connection, project_id: str) -> List['Ontology']:
        if project_id is None:
            raise BaseError('Project ID must be defined!')
        result = con.get('/v2/ontologies/metadata/' + quote_plus(project_id) + '?allLanguages=true')
        return Ontology.allOntologiesFromJsonObj(con, result)

    @staticmethod
    def getOntologyFromServer(con: Connection, shortcode: str, name: str) -> 'Ontology':
        result = con.get("/ontology/" + shortcode + "/" + name + "/v2")
        return Ontology.fromJsonObj(con, result)

    def createDefinitionFileObj(self):
        ontology = {
            "name": self._name,
            "label": self._label,
            "properties": [],
            "resources": []
        }
        if self._comment is not None:
            ontology["comment"] = self._comment
        for prop in self._property_classes:
            if "knora-api:hasLinkToValue" in prop.superproperties:
                self._skiplist.append(self._name + ":" + prop.name)
                continue
            ontology["properties"].append(prop.createDefinitionFileObj(self.context, self._name))

        for res in self._resource_classes:
            ontology["resources"].append(res.createDefinitionFileObj(self.context, self._name, self._skiplist))

        return ontology

    def print(self, short: bool = True) -> None:
        print('Ontology Info:')
        print('  Id:                   {}'.format(self._id))
        print('  Label:                {}'.format(self._label))
        print('  Name:                 {}'.format(self._name))
        print('  Project:              {}'.format(self._project))
        print('  LastModificationDate: {}'.format(str(self._lastModificationDate)))
        if not short:
            print('  Property Classes:')
            if self._property_classes:
                for pc in self._property_classes:
                    pc.print(4)
            print('  Resource Classes:')
            if self._resource_classes:
                for rc in self._resource_classes:
                    rc.print(4)




"""
The code in this file handles the arguments passed by the user from the command line and calls the requested actions.
"""
import argparse
import datetime
import os
import sys

import pkg_resources  # part of setuptools

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

from dsplib.utils.onto_create_lists import create_lists
from dsplib.utils.onto_create_ontology import create_ontology
from dsplib.utils.onto_get import get_ontology
from dsplib.utils.excel_to_json_lists import list_excel2json, validate_list_with_schema
from dsplib.utils.excel_to_json_resources import resources_excel2json
from dsplib.utils.excel_to_json_properties import properties_excel2json
from dsplib.utils.onto_validate import validate_ontology
from dsplib.utils.xml_upload import xml_upload


def program(args: list) -> None:
    """
    The program parses the command line arguments and calls the requested action

    Args:
        args: list of arguments passed by the user from the command line

    Returns:
        None
    """
    version = pkg_resources.require('dsp-tools')[0].version
    now = datetime.datetime.now()

    # parse the arguments of the command line
    parser = argparse.ArgumentParser(
            description=f'dsp-tools (Version {version}) DaSCH Service Platform data modelling tools (© {now.year} by DaSCH).')

    subparsers = parser.add_subparsers(title='Subcommands', description='Valid subcommands are', help='sub-command help')

    parser_create = subparsers.add_parser('create', help='Upload an ontology and/or list(s) from a JSON file to the DaSCH '
                                                         'Service Platform')
    parser_create.set_defaults(action='create')
    parser_create.add_argument('-s', '--server', type=str, default='http://0.0.0.0:3333', help='URL of the DSP server')
    parser_create.add_argument('-u', '--user', default='root@example.com', help='Username for DSP server')
    parser_create.add_argument('-p', '--password', default='test', help='The password for login')
    parser_create.add_argument('-V', '--validate', action='store_true', help='Do only validation of JSON, no upload of the '
                                                                             'ontology')
    parser_create.add_argument('-L', '--listfile', type=str, default='lists.json', help='Name of list node informationfile')
    parser_create.add_argument('-l', '--lists', action='store_true', help='Upload only the list(s)')
    parser_create.add_argument('-v', '--verbose', action='store_true', help='Verbose feedback')
    parser_create.add_argument('-d', '--dump', action='store_true', help='dump test files for DSP-API requests')
    parser_create.add_argument('datamodelfile', help='path to data model file')

    parser_get = subparsers.add_parser('get', help='Get the ontology (data model) of a project from the DaSCH Service Platform.')
    parser_get.set_defaults(action='get')
    parser_get.add_argument('-u', '--user', default='root@example.com', help='Username for DSP server')
    parser_get.add_argument('-p', '--password', default='test', help='The password for login')
    parser_get.add_argument('-s', '--server', type=str, default='http://0.0.0.0:3333', help='URL of the DSP server')
    parser_get.add_argument('-P', '--project', type=str, help='Shortcode, shortname or iri of project', required=True)
    parser_get.add_argument('-v', '--verbose', action='store_true', help='Verbose feedback')
    parser_get.add_argument('datamodelfile', help='Path to the file the ontology should be written to', default='onto.json')

    parser_upload = subparsers.add_parser('xmlupload', help='Upload data from an XML file to the DaSCH Service Platform.')
    parser_upload.set_defaults(action='xmlupload')
    parser_upload.add_argument('-s', '--server', type=str, default='http://0.0.0.0:3333', help='URL of the DSP server')
    parser_upload.add_argument('-u', '--user', type=str, default='root@example.com', help='Username for DSP server')
    parser_upload.add_argument('-p', '--password', type=str, default='test', help='The password for login')
    parser_upload.add_argument('-V', '--validate', action='store_true', help='Do only validation of XML, no upload of the data')
    parser_upload.add_argument('-i', '--imgdir', type=str, default='.', help='Path to folder containing the images')
    parser_upload.add_argument('-S', '--sipi', type=str, default='http://0.0.0.0:1024', help='URL of SIPI server')
    parser_upload.add_argument('-v', '--verbose', action='store_true', help='Verbose feedback')
    parser_upload.add_argument('xmlfile', help='path to xml file containing the data', default='data.xml')

    parser_excel_lists = subparsers.add_parser('excel', help='Create a JSON list from one or multiple Excel files. The JSON '
                                                             'list can be integrated into a JSON ontology. If the list should '
                                                             'contain multiple languages, an Excel file has to be used for '
                                                             'each language. The filenames should contain the language as '
                                                             'label, p.ex. liste_de.xlsx, list_en.xlsx. The language is then '
                                                             'taken from the filename. Only files with extension .xlsx are '
                                                             'considered.')
    parser_excel_lists.set_defaults(action='excel')
    parser_excel_lists.add_argument('-l', '--listname', type=str, help='Name of the list to be created (filename is taken if '
                                                                       'omitted)', default=None)
    parser_excel_lists.add_argument('excelfolder', help='Path to the folder containing the Excel file(s)', default='lists')
    parser_excel_lists.add_argument('outfile', help='Path to the output JSON file containing the list data', default='list.json')

    parser_excel_resources = subparsers.add_parser('excel2resources', help='Create a JSON file from an Excel file containing '
                                                                           'resources for a DSP ontology. ')
    parser_excel_resources.set_defaults(action='excel2resources')
    parser_excel_resources.add_argument('excelfile', help='Path to the Excel file containing the resources',
                                        default='resources.xlsx')
    parser_excel_resources.add_argument('outfile', help='Path to the output JSON file containing the resource data',
                                        default='resources.json')

    parser_excel_properties = subparsers.add_parser('excel2properties', help='Create a JSON file from an Excel file containing '
                                                                             'properties for a DSP ontology. ')
    parser_excel_properties.set_defaults(action='excel2properties')
    parser_excel_properties.add_argument('excelfile', help='Path to the Excel file containing the properties',
                                         default='properties.xlsx')
    parser_excel_properties.add_argument('outfile', help='Path to the output JSON file containing the properties data',
                                         default='properties.json')

    args = parser.parse_args(args)

    if not hasattr(args, 'action'):
        parser.print_help(sys.stderr)
        exit(0)

    if args.action == 'create':
        if args.lists:
            if args.validate:
                validate_list_with_schema(args.datamodelfile)
            else:
                create_lists(input_file=args.datamodelfile,
                             lists_file=args.listfile,
                             server=args.server,
                             user=args.user,
                             password=args.password,
                             verbose=args.verbose,
                             dump=args.dump)
        else:
            if args.validate:
                validate_ontology(args.datamodelfile)
            else:
                create_ontology(input_file=args.datamodelfile,
                                lists_file=args.listfile,
                                server=args.server,
                                user=args.user,
                                password=args.password,
                                verbose=args.verbose,
                                dump=args.dump if args.dump else False)
    elif args.action == 'get':
        get_ontology(projident=args.project,
                     outfile=args.datamodelfile,
                     server=args.server,
                     user=args.user,
                     password=args.password,
                     verbose=args.verbose)
    elif args.action == 'xmlupload':
        xml_upload(input_file=args.xmlfile,
                   server=args.server,
                   user=args.user,
                   password=args.password,
                   imgdir=args.imgdir,
                   sipi=args.sipi,
                   verbose=args.verbose,
                   validate_only=args.validate)
    elif args.action == 'excel':
        list_excel2json(listname=args.listname,
                        excelfolder=args.excelfolder,
                        outfile=args.outfile)
    elif args.action == 'excel2resources':
        resources_excel2json(excelfile=args.excelfile,
                             outfile=args.outfile)
    elif args.action == 'excel2properties':
        properties_excel2json(excelfile=args.excelfile,
                              outfile=args.outfile)


def main():
    """Main entry point of the program as referenced in setup.py"""
    program(sys.argv[1:])


if __name__ == '__main__':
    program(sys.argv[1:])



import unittest

from dsplib.models.langstring import Languages, LangStringParam, LangString


class TestLangString(unittest.TestCase):

    def test_langstring_instantiation1(self):
        """Test a LangString without language."""
        ls = LangString('Ein simpler String')
        self.assertEqual(ls[None], 'Ein simpler String')

    def test_langstring_instantiation2(self):
        """Test a LangString using string and Languages-enums as index."""
        ls = LangString({'de': 'Ein simpler String', 'fr': 'Quelque chose en français'})
        self.assertEqual(ls['de'], 'Ein simpler String')
        self.assertEqual(ls[Languages.FR], 'Quelque chose en français')

    def test_langstring_instantiation3(self):
        """Test a LangString using string and Languages-enums as index."""
        ls = LangString({Languages.DE: 'Ein simpler String', Languages.FR: 'Quelque chose en français'})
        self.assertEqual(ls[Languages.DE], 'Ein simpler String')
        self.assertEqual(ls['fr'], 'Quelque chose en français')

    def test_langstring_change(self):
        """test if changing a LangString item works."""
        ls = LangString({Languages.DE: 'Ein simpler String', Languages.FR: 'Quelque chose en français'})
        ls['de'] = 'gagaga'
        self.assertEqual(ls[Languages.DE], 'gagaga')
        self.assertEqual(ls['fr'], 'Quelque chose en français')

    def test_langstring_fromjson(self):
        """Test reading a LangString from JSON as used in Knora Admin."""
        test = [{
            'language': 'en',
            'value': 'This is a test'
        }, {
            'language': 'de',
            'value': 'Das ist ein Test'
        }]
        ls = LangString.fromJsonObj(test)
        self.assertEqual(ls['de'], 'Das ist ein Test')
        self.assertEqual(ls[Languages.EN], 'This is a test')

    def test_langstring_fromjsonld(self):
        """Test reading a LangString from JSON-LD as used in Knora data/ontologies"""
        test = [{
            '@language': 'en',
            '@value': 'This is a test'
        }, {
            '@language': 'de',
            '@value': 'Das ist ein Test'
        }]
        ls = LangString.fromJsonLdObj(test)
        self.assertEqual(ls['de'], 'Das ist ein Test')
        self.assertEqual(ls[Languages.EN], 'This is a test')

    def test_langstring_tojson(self):
        """Test converting a LangString to JSON and JSON-LD"""
        ls = LangString('Ein simpler String')
        json = ls.toJsonObj()
        self.assertEqual(json, 'Ein simpler String')
        json = ls.toJsonLdObj()
        self.assertEqual(json, 'Ein simpler String')

        ls = LangString({Languages.DE: 'Ein simpler String', Languages.FR: 'Quelque chose en français'})
        json = ls.toJsonObj()
        expected = [{'language': 'de',
                     'value': 'Ein simpler String'},
                    {'language': 'fr',
                     'value': 'Quelque chose en français'}]
        self.assertEqual(json, expected)
        jsonld = ls.toJsonLdObj()
        expected = [{'@language': 'de',
                     '@value': 'Ein simpler String'},
                    {'@language': 'fr',
                     '@value': 'Quelque chose en français'}]
        self.assertEqual(jsonld, expected)

    def test_langstring_emptyness(self):
        """Test if a LanGstring can be emptied and if the emptyness is detected."""
        ls = LangString()
        self.assertTrue(ls.isEmpty())
        ls = LangString('Ein simpler String')
        ls.empty()
        self.assertTrue(ls.isEmpty())
        ls = LangString({Languages.DE: 'Ein simpler String', Languages.FR: 'Quelque chose en français'})
        ls.empty()
        self.assertTrue(ls.isEmpty())

    def test_langstring_iterator(self):
        """Test iterating over a LangString."""
        ls = LangString({Languages.DE: 'Ein simpler String', Languages.FR: 'Quelque chose en français'})
        for tmp in ls:
            if tmp[0] == Languages.DE:
                self.assertEqual(tmp[1], 'Ein simpler String')
            elif tmp[0] == Languages.FR:
                self.assertEqual(tmp[1], 'Quelque chose en français')


if __name__ == '__main__':
    unittest.main()



import unittest

from dsplib.models.connection import Connection
from dsplib.models.helpers import BaseError, Actions, LastModificationDate
from dsplib.models.langstring import Languages, LangStringParam, LangString
from dsplib.models.ontology import Ontology
from dsplib.models.resourceclass import ResourceClass, HasProperty


class TestResourceClass(unittest.TestCase):
    project = "http://rdfh.ch/projects/0001"

    label = LangString({Languages.DE: 'MyResClassLabel'})
    comment = LangString({Languages.DE: 'This is a resource class for testing'})
    name = 'MyResClassName'


    def test_ResourceClass_create(self):
        #
        # Connect to Knora
        #
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')

        #
        # Create a test ontology
        #
        onto = Ontology(
            con=con,
            project=self.project,
            name='resclass-test-onto-1',
            label='resclass test ontology 1',
        ).create()
        last_modification_date = onto.lastModificationDate
        self.assertIsNotNone(onto.id)
        #
        # Create new resource class
        #
        last_modification_date, resclass = ResourceClass(
            con=con,
            context=onto.context,
            name=self.name,
            ontology_id=onto.id,
            label=self.label,
            comment=self.comment
        ).create(last_modification_date)
        onto.lastModificationDate = last_modification_date
        self.assertIsNotNone(resclass.id)

        self.assertEqual(resclass.name, self.name)
        self.assertEqual(resclass.label['de'], self.label['de'])
        self.assertEqual(resclass.comment['de'], self.comment['de'])

        #
        # Delete the new resource class
        #
        last_modification_date = resclass.delete(last_modification_date)
        onto.lastModificationDate = last_modification_date


    def test_ResourceClass_update(self):
        #
        # Connect to Knora
        #
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')

        #
        # Create a test ontology
        #
        onto = Ontology(
            con=con,
            project=self.project,
            name='resclass-test-onto-2',
            label='resclass test ontology 2',
        ).create()
        last_modification_date = onto.lastModificationDate
        self.assertIsNotNone(onto.id)
        #
        # create test resource class
        #
        last_modification_date, resclass = ResourceClass(
            con=con,
            context=onto.context,
            name=self.name,
            ontology_id=onto.id,
            label=self.label,
            comment=self.comment
        ).create(last_modification_date)
        onto.lastModificationDate = last_modification_date
        self.assertIsNotNone(resclass.id)

        #
        # Modify the resource class
        #
        resclass.addLabel('en', "This is english gaga")
        resclass.rmLabel('de')
        resclass.addComment('it', "Commentario italiano")
        last_modification_date, resclass = resclass.update(last_modification_date)
        onto.lastModificationDate = last_modification_date
        self.assertEqual(resclass.label['en'], "This is english gaga")
        self.assertEqual(resclass.comment['it'], "Commentario italiano")

        #
        # Now delete the resource class to clean up
        #
        last_modification_date = resclass.delete(last_modification_date)
        onto.lastModificationDate = last_modification_date


if __name__ == '__main__':
    unittest.main()



import os

import requests

from .helpers import BaseError


def on_api_error(res):
    """
    Checks for any API errors

    Args:
        res: the response from the API which is checked, usually in JSON format

    Returns:
        Knora Error that is being raised
    """

    if res.status_code != 200:
        raise BaseError("SIPI-ERROR: status code=" + str(res.status_code) + "\nMessage:" + res.text)

    if 'error' in res:
        raise BaseError("SIPI-ERROR: API error: " + res.error)


class Sipi:
    """Represents the Sipi instance"""

    def __init__(self, sipi_server: str, token: str):
        self.sipi_server = sipi_server
        self.token = token

    def upload_bitstream(self, filepath):
        """
        Uploads a bitstream to the Sipi server

        Args:
            filepath: path to the file, could be either absolute or relative

        Returns:
            API response
        """
        with open(filepath, 'rb') as bitstream_file:
            files = {'file': (os.path.basename(filepath), bitstream_file), }
            req = requests.post(self.sipi_server + "/upload?token=" + self.token, files=files)
        on_api_error(req)
        print(f'Uploaded file {filepath}')
        res = req.json()
        return res



"""This module handles all the operations which are used for the creation of JSON lists form Excel files."""
import csv
import glob
import json
import os
import re
import unicodedata
from typing import List

import jsonschema
from jsonschema import validate
from openpyxl import load_workbook

list_of_lists = []
cell_names = []


def get_values_from_excel(excelfiles: List[str], base_file: str, parentnode: {}, row: int, col: int, preval: List[str]) -> int:
    """
    This function calls itself recursively to go through the Excel files. It extracts the cell values and creates the JSON list
    file.

    Args:
        base_file: File name of the base file
        excelfiles: List of Excel files with the values in different languages
        parentnode: Name(s) of the parent node(s) of the actual node
        row: The index of the actual row of the Excel sheet
        col: The index of the actual column of the Excel sheet
        preval: List of previous values, needed to check the consistency of the list hierarchy

    Returns:
        int: Row index for the next loop (actual row index minus 1)
    """
    nodes = []
    currentnode = {}
    wb = load_workbook(filename=base_file, read_only=True)
    worksheet = wb.worksheets[0]
    cell = worksheet.cell(column=col, row=row)

    if col > 1:
        # append the cell value of the parent node (which is one value to the left of the actual cell) to the list of previous
        # values
        preval.append(worksheet.cell(column=col - 1, row=row).value)

    while cell.value:
        # check if all predecessors in row (values to the left) are consistent with the values in preval list
        for idx, val in enumerate(preval[:-1]):
            if val != worksheet.cell(column=idx + 1, row=row).value:
                print(f'Inconsistency in Excel list: {val} not equal to {worksheet.cell(column=idx + 1, row=row).value}')
                quit()

        # loop through the row until the last (furthest right) value is found
        if worksheet.cell(column=col + 1, row=row).value:
            row = get_values_from_excel(excelfiles=excelfiles, base_file=base_file, parentnode=currentnode, col=col + 1, row=row,
                                        preval=preval)

        # if value was last in row (no further values to the right), it's a node, continue here
        else:
            # check if there are duplicate nodes (i.e. identical rows), quit the program if so
            new_check_list = preval.copy()
            new_check_list.append(cell.value)

            list_of_lists.append(new_check_list)

            if check_list_for_duplicates(list_of_lists):
                print('There is at least one duplicate node in the list. Found duplicate:', cell.value)
                quit()

            # create a simplified version of the cell value and use it as name of the node
            cellname = simplify_name(cell.value)
            cell_names.append(cellname)

            # append a number (p.ex. node-name-2) if there are list nodes with identical names
            if check_list_for_duplicates(cell_names):
                n = cell_names.count(cellname)
                if n > 1:
                    cellname = cellname + '-' + str(n)

            labels_dict = {}

            # read label values from the other Excel files (other languages)
            for filename_other_lang in excelfiles:
                wb_other_lang = load_workbook(filename=filename_other_lang, read_only=True)
                ws_other_lang = wb_other_lang.worksheets[0]

                lang = os.path.splitext(filename_other_lang)[0].split('_')[-1]
                labels_dict[lang] = ws_other_lang.cell(column=col, row=row).value

            # create current node from extracted cell values and append it to the nodes list
            currentnode = {'name': cellname, 'labels': labels_dict}

            nodes.append(currentnode)

            print(f'Added list node: {cell.value} ({cellname})')

        # go one row down and repeat loop if there is a value
        row += 1
        cell = worksheet.cell(column=col, row=row)

    if col > 1:
        preval.pop()

    # add the new nodes to the parentnode
    parentnode['nodes'] = nodes

    return row - 1


def make_json_list_from_excel(rootnode: {}, excelfiles: List[str]) -> None:
    """
    Reads Excel files and makes a JSON list file from them. The JSON can then be used in an ontology that is uploaded to the
    DaSCH Service Platform.

    Args:
        rootnode: The root node of the JSON list
        excelfiles: A list with all the Excel files to be processed

    Returns:
        None

    """
    # Define starting point in Excel file
    startrow = 1
    startcol = 1

    # Check if English file is available and take it as base file, take last one from list of Excel files if English is not
    # available. The node names are later  derived from the labels of the base file.
    base_file = ''

    for filename in excelfiles:
        base_file = filename
        if '_en.xlsx' in os.path.basename(filename):
            base_file = filename
            break

    get_values_from_excel(excelfiles=excelfiles, base_file=base_file, parentnode=rootnode, row=startrow, col=startcol, preval=[])


def check_list_for_duplicates(list_to_check: list) -> bool:
    """
    Checks if the given list contains any duplicate items.

    Args:
        list_to_check: A list of items to be checked for duplicates

    Returns:
        True if there is a duplicate, false otherwise

    """
    has_duplicates = False

    for item in list_to_check:
        if list_to_check.count(item) > 1:
            has_duplicates = True

    return has_duplicates


def simplify_name(value: str) -> str:
    """
    Simplifies a given value in order to use it as node name

    Args:
        value: The value to be simplified

    Returns:
        str: The simplified value

    """
    simplified_value = str(value).lower()

    # normalize characters (p.ex. ä becomes a)
    simplified_value = unicodedata.normalize('NFKD', simplified_value)

    # replace forward slash and whitespace with a dash
    simplified_value = re.sub('[/\\s]+', '-', simplified_value)

    # delete all characters which are not letters, numbers or dashes
    simplified_value = re.sub('[^A-Za-z0-9\\-]+', '', simplified_value)

    return simplified_value


def check_language_code(lang_code: str) -> bool:
    """
    Checks if a given language code is valid. The code is valid if it is listed in language-codes-3b2_csv.csv. This
    file provides all ISO 639-1 and ISO 639-2 language codes.

    Args:
        lang_code: the language code to be checked

    Returns:
        True if valid, False if not

    """
    current_dir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(current_dir, 'language-codes-3b2_csv.csv'), 'r') as language_codes_file:
        language_codes = csv.reader(language_codes_file, delimiter=',')
        for row in language_codes:
            if lang_code in row:
                return True
    return False


def make_root_node_from_args(excelfiles: List[str], listname_from_args: str) -> dict:
    """
    Creates the root node for the JSON list

    Args:
        excelfiles: List of Excel files (names) to be checked
        listname_from_args: Listname from arguments provided by the user via the command line

    Returns:
        dict: The root node of the list as dictionary (JSON)

    """
    rootnode_labels_dict = {}
    listname = listname_from_args
    listname_en = ''

    for filename in excelfiles:
        basename = os.path.basename(filename)
        label, lang_code = os.path.splitext(basename)[0].rsplit('_', 1)

        # check if language code is valid
        if not check_language_code(lang_code):
            print('Invalid language code is used. Only language codes from ISO 639-1 and ISO 639-2 are accepted.')
            quit()

        rootnode_labels_dict[lang_code] = label
        listname = label

        if '_en.xlsx' in filename:
            listname_en = label

    # if an english list is available use its label as listname
    if listname_en:
        listname = listname_en

    # if the user provided a listname use it
    if listname_from_args:
        listname = listname_from_args

    rootnode = {'name': listname, 'labels': rootnode_labels_dict}

    return rootnode


def validate_list_with_schema(json_list: str) -> bool:
    """
    This function checks if a list is valid according to the schema.

    Args:
        json_list (json): the json list to be validated

    Returns:
        True if the list passed validation, False otherwise

    """
    current_dir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(current_dir, '../schemas/lists-only.json')) as schema:
        list_schema = json.load(schema)

    try:
        validate(instance=json_list, schema=list_schema)
    except jsonschema.exceptions.ValidationError as err:
        print(err)
        return False
    print('List passed schema validation.')
    return True


def prepare_list_creation(excelfolder: str, listname: str):
    """
    Gets the excelfolder parameter and checks the validity of the files. It then makes the root node for the list.

    Args:
        excelfolder: path to the folder containing the Excel file(s)
        listname: name of the list to be created

    Returns:
        rootnode (dict): The rootnode of the list as a dictionary
        excel_files (List[str]): list of the Excel files to process
    """
    # reset the global variables before list creation starts
    global cell_names
    global list_of_lists

    list_of_lists = []
    cell_names = []

    # check if the given folder parameter is actually a folder
    if not os.path.isdir(excelfolder):
        print(excelfolder, 'is not a directory.')
        exit()

    # create a list with all excel files from the path provided by the user
    excel_files = [filename for filename in glob.iglob(f'{excelfolder}/*.xlsx') if
                   not os.path.basename(filename).startswith('~$')]

    # check if all excel_files are actually files
    print('Found the following files:')
    for file in excel_files:
        print(file)
        if not os.path.isfile(file):
            print(file, 'is not a valid file.')
            exit()

    # create root node of list
    rootnode = make_root_node_from_args(excel_files, listname)

    return rootnode, excel_files


def list_excel2json(listname: str, excelfolder: str, outfile: str):
    """
    Takes the arguments from the command line, checks folder and files and starts the process of list creation.

    Args:
        listname: name of the list to be created, file name is taken if omitted
        excelfolder: path to the folder containing the Excel file(s)
        outfile: path to the JSON file the output is written into

    Return:
        None
    """
    # get the Excel files from the folder and crate the rootnode of the list
    rootnode, excel_files = prepare_list_creation(excelfolder, listname)

    # create the list from the Excel files
    make_json_list_from_excel(rootnode, excel_files)

    # validate created list with schema
    if validate_list_with_schema(json.loads(json.dumps(rootnode, indent=4))):
        # write final list to JSON file if list passed validation
        with open(outfile, 'w', encoding='utf-8') as fp:
            json.dump(rootnode, fp, indent=4, sort_keys=False, ensure_ascii=False)
            print('List was created successfully and written to file:', outfile)
    else:
        print('List is not valid according to schema.')



import json
from typing import List

from .expand_all_lists import expand_lists_from_excel
from .onto_validate import validate_ontology
from ..models.connection import Connection
from ..models.listnode import ListNode
from ..models.project import Project


def list_creator(con: Connection, project: Project, parent_node: ListNode, nodes: List[dict]):
    """
    Creates the list on the DSP server

    Args:
        con: The connection to the DSP server
        project: The project which the lists should be added
        parent_node: The root node of the list
        nodes: List of nodes the list is made of

    Returns:
        The list of all nodes with their names and respective IDs
    """
    nodelist = []
    for node in nodes:
        new_node = ListNode(con=con, project=project, label=node["labels"], comments=node.get("comments"), name=node["name"],
                            parent=parent_node).create()
        if node.get('nodes') is not None:
            subnode_list = list_creator(con, project, new_node, node['nodes'])
            nodelist.append({new_node.name: {"id": new_node.id, 'nodes': subnode_list}})
        else:
            nodelist.append({new_node.name: {"id": new_node.id}})
    return nodelist


def create_lists(input_file: str, lists_file: str, server: str, user: str, password: str, verbose: bool, dump: bool = False):
    """
    Creates the lists on the DSP server

    Args:
        input_file: Path to the json data model file
        lists_file: Output file for the list node names and their respective IRI
        server: URL of the DSP server
        user: Username (e-mail) for the DSP server, has to have the permissions to create an ontology
        password: Password of the user
        verbose: Verbose output if True
        dump: ???

    Returns:
        list_root_nodes: Dictionary of node names and their respective IRI
    """
    # read the ontology from the input file
    with open(input_file) as f:
        onto_json_str = f.read()

    data_model = json.loads(onto_json_str)

    # expand all lists referenced in the list section of the data model
    new_lists = expand_lists_from_excel(data_model)

    # add the newly created lists from Excel to the ontology
    data_model['project']['lists'] = new_lists

    # validate the ontology
    if validate_ontology(data_model):
        pass
    else:
        quit()

    # Connect to the DaSCH Service Platform API
    con = Connection(server)
    con.login(user, password)

    if dump:
        con.start_logging()

    # get the project which must exist
    project = Project(con=con, shortcode=data_model['project']['shortcode'], ).read()
    assert project is not None

    # create the lists
    if verbose:
        print('Create lists...')

    lists = data_model['project'].get('lists')
    list_root_nodes = {}
    if lists is not None:
        for rootnode in lists:
            if verbose:
                print('  Create list:' + rootnode['name'])
            root_list_node = ListNode(con=con, project=project, label=rootnode['labels'],  # comment=rootnode.get('comments'),
                                      name=rootnode['name']).create()
            if rootnode.get('nodes') is not None:
                list_nodes = list_creator(con, project, root_list_node, rootnode['nodes'])
                list_root_nodes[rootnode['name']] = {'id': root_list_node.id, 'nodes': list_nodes}

    with open(lists_file, 'w', encoding='utf-8') as fp:
        json.dump(list_root_nodes, fp, indent=3, sort_keys=True)
        print(f'The IRI for the created nodes can be found in {lists_file}.')

    return list_root_nodes



import json
import os
from typing import Dict, Union

import jsonschema
from jsonschema import validate

from ..utils.expand_all_lists import expand_lists_from_excel


def validate_ontology(input_file_or_json: Union[str, Dict, os.PathLike]) -> bool:
    """
    Validates an ontology against the knora schema

    Args:
        input_file_or_json: the ontology to be validated, can either be a file or a json string (dict)

    Returns:
        True if ontology passed validation, False otherwise
    """
    data_model = ''

    if isinstance(input_file_or_json, dict):
        data_model = input_file_or_json
    elif os.path.isfile(input_file_or_json):
        with open(input_file_or_json) as f:
            onto_json_str = f.read()
        data_model = json.loads(onto_json_str)
    else:
        print('Input is not valid.')
        quit()

    # expand all lists referenced in the list section of the data model
    new_lists = expand_lists_from_excel(data_model)

    # add the newly created lists from Excel to the ontology
    data_model['project']['lists'] = new_lists

    # validate the data model against the schema
    current_dir = os.path.dirname(os.path.realpath(__file__))

    with open(os.path.join(current_dir, '../schemas/ontology.json')) as s:
        schema = json.load(s)

    try:
        validate(instance=data_model, schema=schema)
    except jsonschema.exceptions.ValidationError as err:
        print(err)
        return False
    print('Data model is syntactically correct and passed validation.')
    return True



"""This test class tests the basic functionalities of dsp-tools"""
import json
import unittest

from knora.dsplib.utils import excel_to_json_lists
from knora.dsplib.utils.excel_to_json_lists import list_excel2json
from knora.dsplib.utils.excel_to_json_properties import properties_excel2json
from knora.dsplib.utils.excel_to_json_resources import resources_excel2json
from knora.dsplib.utils.onto_create_ontology import create_ontology
from knora.dsplib.utils.onto_get import get_ontology
from knora.dsplib.utils.onto_validate import validate_ontology
from knora.dsplib.utils.xml_upload import xml_upload


class TestTools(unittest.TestCase):
    def setUp(self) -> None:
        """Is executed before each test"""
        excel_to_json_lists.list_of_lists = []
        excel_to_json_lists.cell_names = []

    def tearDown(self) -> None:
        """Is executed after each test"""
        excel_to_json_lists.list_of_lists = []
        excel_to_json_lists.cell_names = []

    def test_get(self):
        global other_tree_list, other_tree_list_out, not_used_list, not_used_list_out, tree_list_root, tree_list_root_out

        with open('testdata/anything-onto.json') as f:
            onto_json_str = f.read()
        anything_onto = json.loads(onto_json_str)

        get_ontology(projident='anything',
                     outfile='_anything-onto.json',
                     server='http://0.0.0.0:3333',
                     user='root@example.com',
                     password='test',
                     verbose=True)

        with open('_anything-onto.json') as f:
            onto_json_str = f.read()
        anything_onto_out = json.loads(onto_json_str)

        self.assertEqual(anything_onto['project']['shortcode'], anything_onto_out['project']['shortcode'])
        self.assertEqual(anything_onto['project']['shortname'], anything_onto_out['project']['shortname'])
        self.assertEqual(anything_onto['project']['longname'], anything_onto_out['project']['longname'])

        for list in anything_onto['project']['lists']:
            list_name = list.get('name')
            if list_name == 'otherTreeList':
                other_tree_list = list
            elif list_name == 'notUsedList':
                not_used_list = list
            elif list_name == 'treelistroot':
                tree_list_root = list

        for list in anything_onto_out['project']['lists']:
            list_name = list.get('name')
            print(list.get('name'))
            if list_name == 'otherTreeList':
                other_tree_list_out = list
            elif list_name == 'notUsedList':
                not_used_list_out = list
            elif list_name == 'treelistroot':
                tree_list_root_out = list

        self.assertEqual(other_tree_list.get('labels'), other_tree_list_out.get('labels'))
        self.assertEqual(other_tree_list.get('comments'), other_tree_list_out.get('comments'))
        self.assertEqual(other_tree_list.get('nodes'), other_tree_list_out.get('nodes'))

        self.assertEqual(not_used_list.get('labels'), not_used_list_out.get('labels'))
        self.assertEqual(not_used_list.get('comments'), not_used_list_out.get('comments'))
        self.assertEqual(not_used_list.get('nodes'), not_used_list_out.get('nodes'))

        self.assertEqual(tree_list_root.get('labels'), tree_list_root_out.get('labels'))
        self.assertEqual(tree_list_root.get('comments'), tree_list_root_out.get('comments'))
        self.assertEqual(tree_list_root.get('nodes'), tree_list_root_out.get('nodes'))

        # TODO fix this test
        # self.assertEqual(anything_onto['project']['ontologies'], anything_onto_out['project']['ontologies'])

    def test_excel_to_json_list(self):
        list_excel2json(listname='my_test_list',
                        excelfolder='testdata/lists',
                        outfile='_lists-out.json')

    def test_excel_to_json_resources(self):
        resources_excel2json(excelfile='testdata/Resources.xlsx',
                             outfile='_resources-out.json')

    def test_excel_to_json_properties(self):
        properties_excel2json(excelfile='testdata/Properties.xlsx',
                              outfile='_properties-out.json')

    def test_validate_ontology(self):
        validate_ontology('testdata/test-onto.json')

    def test_create_ontology(self):
        create_ontology(input_file='testdata/test-onto.json',
                        lists_file='lists-out.json',
                        server='http://0.0.0.0:3333',
                        user='root@example.com',
                        password='test',
                        verbose=True,
                        dump=True)

    def test_xml_upload(self):
        xml_upload(input_file='testdata/test-data.xml',
                   server='http://0.0.0.0:3333',
                   user='root@example.com',
                   password='test',
                   imgdir='testdata/bitstreams',
                   sipi='http://0.0.0.0:1024',
                   verbose=True,
                   validate_only=False)


if __name__ == '__main__':
    unittest.main()



import json
import re
from typing import Dict

from ..models.connection import Connection
from ..models.listnode import ListNode
from ..models.ontology import Ontology
from ..models.project import Project


def get_ontology(projident: str, outfile: str, server: str, user: str, password: str, verbose: bool) -> bool:
    con = Connection(server)
    # con.login(user, password)
    if re.match("^[0-9aAbBcCdDeEfF]{4}$", projident):
        project = Project(con=con, shortcode=projident)
    elif re.match("^[\\w-]+$", projident):
        project = Project(con=con, shortname=projident)
    elif re.match("^(http)s?://([\\w\\.\\-~]+:?\\d{,4})(/[\\w\\-~]+)+$", projident):
        project = Project(con=con, shortname=projident)
    else:
        print("Invalid project identification!")
        return False

    project = project.read()

    projectobj = project.createDefinitionFileObj()

    #
    # now collect the lists
    #
    listroots = ListNode.getAllLists(con=con, project_iri=project.id)
    listobj = []
    for listroot in listroots:
        complete_list = listroot.getAllNodes()
        listobj.append(complete_list.createDefinitionFileObj())
    projectobj["lists"] = listobj

    projectobj["ontologies"] = []
    prefixes: Dict[str, str] = {}
    ontologies = Ontology.getProjectOntologies(con, project.id)
    ontology_ids = [x.id for x in ontologies]
    for ontology in ontology_ids:
        oparts = ontology.split("/")
        name = oparts[len(oparts) - 2]
        shortcode = oparts[len(oparts) - 3]
        ontology = Ontology.getOntologyFromServer(con=con, shortcode=shortcode, name=name)
        projectobj["ontologies"].append(ontology.createDefinitionFileObj())
        prefixes.update(ontology.context.get_externals_used())

    umbrella = {"prefixes": prefixes, "project": projectobj}

    with open(outfile, 'w', encoding='utf8') as outfile:
        json.dump(umbrella, outfile, indent=3, ensure_ascii=False)



import requests
import json
from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union

from .helpers import Actions, BaseError

@strict
class Connection:
    """
    An Connection instance represents a connection to a Knora server.

    Attributes
    ----------

    none (internal use attributes should not be modified/set directly)
    """

    _server: str
    _prefixes: Union[Dict[str, str], None]
    _token: Union[str, None]
    _log: bool

    def __init__(self, server: str, prefixes: Dict[str, str] = None):
        """
        Constructor requiring the server address, the user and password of KNORA
        :param server: Address of the server, e.g https://api.dasch.swiss
        :param prefixes: Ontology prefixes used
        """

        self._server = server
        self._prefixes = prefixes
        self._token = None
        self._log = False

    def login(self, email: str, password: str) -> None:
        """
        Method to login into KNORA which creates a session token.
        :param email: Email of user, e.g., root@example.com
        :param password: Password of the user, e.g. test
        """

        credentials = {
            "email": email,
            "password": password
        }
        jsondata = json.dumps(credentials)

        req = requests.post(
            self._server + '/v2/authentication',
            headers={'Content-Type': 'application/json; charset=UTF-8'},
            data=jsondata
        )
        self.on_api_error(req)
        result = req.json()
        self._token = result["token"]

    def get_token(self) -> str:
        """
        Returns the token
        :return: token string
        """

        return self._token

    @property
    def token(self) -> str:
        return self._token

    def start_logging(self):
        self._log = True

    def stop_logging(self):
        self._log = False

    def logout(self) -> None:
        """
        Performs a logout
        :return: None
        """

        if self._token is not None:
            req = requests.delete(
                self._server + '/v2/authentication',
                headers={'Authorization': 'Bearer ' + self._token}
            )
            self.on_api_error(req)
            self._token = None

    def __del__(self):
        pass
        #self.logout()

    def on_api_error(self, res) -> None:
        """
        Method to check for any API errors
        :param res: The input to check, usually JSON format
        :return: Possible Error that is being raised
        """

        if res.status_code != 200:
            raise BaseError("KNORA-ERROR: status code=" + str(res.status_code) + "\nMessage:" + res.text)

        if 'error' in res:
            raise BaseError("KNORA-ERROR: API error: " + res.error)

    def post(self, path: str, jsondata: Optional[str] = None):
        """
        Post Json data to a given server using a HTTP POST request
        :param path: Path of RESTful route
        :param jsondata: Valid JSON as string
        :return: Response from server
        """

        if path[0] != '/':
            path = '/' + path
        headers = None
        if jsondata is None:
            if self._token is not None:
                headers = {'Authorization': 'Bearer ' + self._token}
                req = requests.post(self._server + path,
                                    headers=headers)
            else:
                req = requests.post(self._server + path)
        else:
            if self._token is not None:
                headers = {'Content-Type': 'application/json; charset=UTF-8',
                           'Authorization': 'Bearer ' + self._token}
                req = requests.post(self._server + path,
                                    headers=headers,
                                    data=jsondata)
            else:
                headers = {'Content-Type': 'application/json; charset=UTF-8'}
                req = requests.post(self._server + path,
                                    headers=headers,
                                    data=jsondata)
        if self._log:
            if jsondata:
                jsonobj = json.loads(jsondata)
            else:
                jsonobj = None
            logobj = {
                "method": "POST",
                "headers": headers,
                "route": path,
                "body": jsonobj,
                "return-headers": dict(req.headers),
                "return": req.json() if req.status_code == 200 else {"status": str(req.status_code),
                                                                     "message": req.text}
            }
            tmp = path.split('/')
            filename = "POST" + "_".join(tmp) + ".json"
            with open(filename, 'w') as f:
                json.dump(logobj, f, indent=3)
        self.on_api_error(req)
        result = req.json()
        return result

    def get(self, path: str, headers: Optional[Dict[str, str]] = None):
        """
        Get data from a server using a HTTP GET request
        :param path: Path of RESTful route
        :param headers: ...
        :return: Response from server
        """

        if path[0] != '/':
            path = '/' + path
        if self._token is None:
            if headers is None:
                req = requests.get(self._server + path)
            else:
                req = requests.get(self._server + path, headers)
        else:
            if headers is None:
                req = requests.get(self._server + path,
                                   headers={'Authorization': 'Bearer ' + self._token})
            else:
                headers['Authorization'] = 'Bearer ' + self._token
                req = requests.get(self._server + path, headers)

        self.on_api_error(req)
        result = req.json()
        return result

    def put(self, path: str, jsondata: Optional[str] = None, content_type: str = 'application/json'):
        """
        Send data to a RESTful server using a HTTP PUT request
        :param path: Path of RESTful route
        :param jsondata: Valid JSON as string
        :param content_type: HTTP Content-Type [default: 'application/json']
        :return:
        """

        if path[0] != '/':
            path = '/' + path
        if jsondata is None:
            req = requests.put(self._server + path,
                               headers={'Authorization': 'Bearer ' + self._token})
        else:
            req = requests.put(self._server + path,
                               headers={'Content-Type': content_type + '; charset=UTF-8',
                                        'Authorization': 'Bearer ' + self._token},
                               data=jsondata)
        self.on_api_error(req)
        result = req.json()
        return result

    def delete(self, path: str, params: Optional[any] = None):
        """
        Send a delete request using the HTTP DELETE request
        :param path: Path of RESTful route
        :return: Response from server
        """

        if path[0] != '/':
            path = '/' + path
        if params is not None:
            req = requests.delete(self._server + path,
                                  headers={'Authorization': 'Bearer ' + self._token},
                                  params=params)

        else:
            req = requests.delete(self._server + path,
                                  headers={'Authorization': 'Bearer ' + self._token})
        self.on_api_error(req)
        result = req.json()
        return result

    def reset_triplestore_content(self):
        rdfdata = [
            {
                "path": "./knora-ontologies/knora-admin.ttl",
                "name": "http://www.knora.org/ontology/knora-admin"
            },
            {
                "path": "./knora-ontologies/knora-base.ttl",
                "name": "http://www.knora.org/ontology/knora-base"
            },
            {
                "path": "./knora-ontologies/standoff-onto.ttl",
                "name": "http://www.knora.org/ontology/standoff"
            },
            {
                "path": "./knora-ontologies/standoff-data.ttl",
                "name": "http://www.knora.org/data/standoff"
            },
            {
                "path": "./knora-ontologies/salsah-gui.ttl",
                "name": "http://www.knora.org/ontology/salsah-gui"
            },
            {
                "path": "./_test_data/all_data/admin-data.ttl",
                "name": "http://www.knora.org/data/admin"
            },
            {
                "path": "./_test_data/all_data/permissions-data.ttl",
                "name": "http://www.knora.org/data/permissions"
            },
            {
                "path": "./_test_data/all_data/system-data.ttl",
                "name": "http://www.knora.org/data/0000/SystemProject"
            }
        ]
        jsondata = json.dumps(rdfdata)
        url = self._server + '/admin/store/ResetTriplestoreContent?prependdefaults=false'

        req = requests.post(url,
                            headers={'Content-Type': 'application/json; charset=UTF-8'},
                            data=jsondata)
        self.on_api_error(req)
        res = req.json()
        #  pprint(res)
        return res


